from enum import Enum, unique
import asyncio


@unique
class QueueName(Enum):
    CONNECTIONS = "connections"
    MESSAGES = "messages"
    SERVERS = "servers"


class QueueManager:
    _queues: dict[QueueName, asyncio.Queue] = {}

    @classmethod
    def get_queue(cls, name: QueueName, maxsize: int = 0) -> asyncio.Queue:
        if not isinstance(name, QueueName):
            raise ValueError(f"Queue name invalid. Available names: {list(QueueName)}")

        if name not in cls._queues:
            cls._queues[name] = asyncio.Queue(maxsize=maxsize)
        return cls._queues[name]


get_queue = QueueManager.get_queue



import asyncio
from xml.etree import ElementTree as ET

from pyjabber import AppConfig
from pyjabber.network.parsers.XMLParser import XMLParser
from pyjabber.queues.QueueManager import get_queue, QueueName
from pyjabber.stream.handlers.ServerStanzaHandler import ServerStanzaHandler
from pyjabber.stream.utils import Stream
from pyjabber.stream.negotiators.ServerStreamNegotiator import StreamServerNegotiator
from pyjabber.utils import ClarkNotation as CN


class XMLParserServerOutgoing(XMLParser):
    def __init__(self, transport, protocol, host):
        super().__init__(transport, protocol, StreamServerNegotiator, ServerStanzaHandler)
        self._connection_queue: asyncio.Queue = get_queue(QueueName.CONNECTIONS)
        self._host = host

        self._initial_stream()

    def startElementNS(self, name, qname, attrs):
        elem = ET.Element(
            CN.clark_from_tuple(name),
            attrib={CN.clark_from_tuple(key): item for key, item in dict(attrs).items()}
        )

        self._stack.append(elem)

    def reset_stack(self) -> None:
        super().reset_stack()
        self._initial_stream()

    def _initial_stream(self):
        initial_stream = Stream.Stream(
            from_=AppConfig.app_config.host,
            to=self._host,
            xmlns=Stream.Namespaces.SERVER.value
        )

        initial_stream = initial_stream.open_tag()
        self._transport.write(initial_stream)



import xml.etree.ElementTree as ET

from sqlalchemy import and_, delete, insert, select, update

from pyjabber import AppConfig
from pyjabber.db.database import DB
from pyjabber.db.model import Model
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID


class Roster:
    """
        Roster plugin.

        Manages the roster list for each registered user in the protocols.

        A roster in XMPP is a protocols-stored contact list that manages contacts, presence status,
        and subscription requests.
        It enables real-time presence updates, contact organization, and synchronization across devices,
        ensuring seamless and private communication.
    """
    __slots__ = ('_handlers', '_roster_in_memory', '_initial_update')

    def __init__(self) -> None:
        self._handlers = {
            "get": self.handle_get,
            "set": self.handle_set,
            "result": self.handle_result
        }

        self._roster_in_memory = {}
        self._initial_update = False

    async def feed(self, jid: JID, element: ET.Element):
        if not self._initial_update:
            await self._update_roster()
        if len(element) != 1:
            return SE.invalid_xml()

        return await self._handlers[element.attrib.get("type")](jid, element)

    async def handle_get(self, jid: JID, element: ET.Element):
        if jid.domain == AppConfig.app_config.host:
            jid = jid.user
        else:
            jid = jid.bare()

        roster = self._roster_in_memory.get(jid)

        iq = IQ(type_=IQ.TYPE.RESULT, id_=element.attrib.get("id"))
        query = ET.SubElement(iq, "{jabber:iq:roster}query")

        for item in roster or []:
            query.append(ET.fromstring(item.get("item")))

        return ET.tostring(iq)

    async def handle_set(self, jid: JID, element: ET.Element):
        query = element.find("{jabber:iq:roster}query")
        if query is None:
            return SE.invalid_xml()

        jid = jid.user if jid.domain == AppConfig.app_config.host else jid.bare()
        new_item = query.findall("{jabber:iq:roster}item")

        if len(new_item) != 1:
            return SE.invalid_xml()

        new_item = new_item[0]

        roster = self._roster_in_memory.get(jid)
        if roster:
            match_item = [i for i in roster if ET.fromstring(i.get("item")).get("jid") == new_item.attrib.get("jid")]
            if match_item:  # UPDATE EXISTING ENTRY
                match_item = match_item[0]
                if new_item.attrib.get("remove") == "remove":  # DELETE ENTRY
                    async with await DB.connection_async() as con:
                        query = delete(Model.Roster).where(
                            and_(
                                Model.Roster.c.jid == jid,
                                Model.Roster.c.roster_item == ET.tostring(match_item.get("item")).decode()
                            )
                        )
                        await con.execute(query)
                        if not AppConfig.app_config.database_in_memory:
                            await con.commit()

                else:  # UPDATE FIELDS OF ENTRY
                    async with await DB.connection_async() as con:
                        query = update(Model.Roster).where(
                            and_(
                                Model.Roster.c.jid == jid,
                                Model.Roster.c.roster_item == match_item.get("item")
                            )
                        ).values({"roster_item": ET.tostring(new_item).decode()})
                        await con.execute(query)
                        if not AppConfig.app_config.database_in_memory:
                            await con.commit()

            else:  # CREATE NEW ENTRY
                if new_item.attrib.get("remove") != "remove":
                    async with await DB.connection_async() as con:
                        query = insert(Model.Roster).values({
                            "jid": jid,
                            "roster_item": ET.tostring(new_item).decode()
                        })
                        await con.execute(query)
                        if not AppConfig.app_config.database_in_memory:
                            await con.commit()

        else:
            async with await DB.connection_async() as con:
                query = insert(Model.Roster).values({
                    "jid": jid,
                    "roster_item": ET.tostring(new_item).decode()
                })
                await con.execute(query)
                if not AppConfig.app_config.database_in_memory:
                    await con.commit()

        await self._update_roster()
        res = IQ(
            id_=element.attrib.get("id"),
            type_=IQ.TYPE.RESULT
        )
        return ET.tostring(res)

    def handle_result(self, _, __):
        # It's safe to ignore this stanza
        return

    async def create_roster_entry(self, jid: JID,  to: JID):
        iq = IQ(
            from_=str(jid),
            type_=IQ.TYPE.SET
        )
        query = ET.SubElement(iq, "{jabber:iq:roster}query")
        if to.domain == AppConfig.app_config.host:
            ET.SubElement(query, "{jabber:iq:roster}item", attrib={"jid": to.user, "subscription": "none"})
        else:
            ET.SubElement(query, "{jabber:iq:roster}item", attrib={"jid": to.bare(), "subscription": "none"})

        return await self.feed(jid, iq)

    @staticmethod
    async def store_pending_sub(to_: str, item: ET.Element) -> None:
        async with await DB.connection_async() as con:
            query = insert(Model.PendingSubs).values({
                "jid": to_,
                "item":  ET.tostring(item).decode()
            })
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

    async def update_item(self, item: ET.Element, id_: int):
        async with await DB.connection_async() as con:
            query = update(Model.Roster).where(Model.Roster.c.id == id_).values({
                "roster_item": ET.tostring(item).decode()
            })
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        await self._update_roster()

    async def _update_roster(self):
        async with await DB.connection_async() as con:
            query = select(
                Model.Roster.c.id,
                Model.Roster.c.jid,
                Model.Roster.c.roster_item
            )
            res = await con.execute(query)
            res = res.fetchall()

        self._roster_in_memory.clear()
        for id_, jid, item in res:
            if jid not in self._roster_in_memory:
                self._roster_in_memory[jid] = []
            self._roster_in_memory[jid].append({"id": id_, "item": item})

    def roster_by_jid(self, jid: JID):
        if jid.domain == AppConfig.app_config.host:
            return self._roster_in_memory.get(jid.user) or []
        return self._roster_in_memory.get(jid.bare()) or []



from xml import sax

from loguru import logger

from pyjabber import AppConfig
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.parsers.XMLParserServerOutgoing import XMLParserServerOutgoing
from pyjabber.network.utils.TransportProxy import TransportProxy
from pyjabber.network.protocols.XMLProtocol import XMLProtocol


class XMLProtocolServerOutgoing(XMLProtocol):
    def __init__(self, namespace, host, connection_timeout):
        super().__init__(namespace, connection_timeout)
        self._host = host

    def connection_made(self, transport):
        """
        Called when a client or another protocols opens a TCP connection to the protocols

        :param transport: The transport object for the connection
        :type transport: asyncio.Transport
        """
        if transport:
            self._peer = transport.get_extra_info('peername')
            logger.info(f"Connection {self._peer}")

            if AppConfig.app_config.verbose:
                self._transport = TransportProxy(transport, self._peer)
            else:
                self._transport = transport

            self._xml_parser = sax.make_parser()
            self._xml_parser.setFeature(sax.handler.feature_namespaces, True)
            self._xml_parser.setFeature(sax.handler.feature_external_ges, False)
            self._xml_parser.setContentHandler(
                XMLParserServerOutgoing(self._transport, self, self._host)
            )

            if self._connection_timeout:
                self._timeout_monitor = StreamAlivenessMonitor(
                    timeout=self._connection_timeout,
                    callback=self.connection_timeout
                )

            self._connection_manager.connection_server(
                self._peer, self._transport, self._host
            )
        else:
            logger.error("Invalid transport")

    def connection_lost(self, exc):
        """
        Called when a protocols closes a TCP connection to the protocols

        :param exc: Exception that caused the connection to close
        :type exc: Exception
        """
        if self._timeout_flag:
            return

        logger.info(f"Connection lost {self._peer}: Reason {exc}")

        self._transport = None
        self._xml_parser = None
        self._timeout_monitor.cancel()

        self._connection_manager.disconnection_server(self._peer)



import json
import os
import pickle
import socket
from dataclasses import dataclass, field, replace
from typing import Dict, List


@dataclass(frozen=True, slots=True)
class Parameters:
    host: str = "localhost"
    client_port: int = 5222
    server_port: int = 5269
    family: socket.AddressFamily = socket.AF_INET
    connection_timeout: int = 60
    database_path: str = os.path.join(os.getcwd(), "pyjabber.db")
    database_purge: bool = False
    database_in_memory: bool = False
    database_debug: bool = False
    cert_path: str = None
    message_persistence: bool = True
    verbose: bool = False
    plugins: List[str] = field(default_factory=lambda: [
            'http://jabber.org/protocol/disco#info',
            'http://jabber.org/protocol/disco#items',
            'http://jabber.org/protocol/pubsub',
            'http://jabber.org/protocol/pubsub#publish',
            'http://jabber.org/protocol/pubsub#subscribe',
            'http://jabber.org/protocol/pubsub#config-node',
            'http://jabber.org/protocol/pubsub#create-nodes',
            'http://jabber.org/protocol/pubsub#delete-nodes',
            'jabber:iq:register',
            'jabber:x:data',
            'urn:xmpp:ping',
            'jabber:iq:rpc',
            'urn:xmpp:http:upload:0'
    ])

    items: Dict[str, Dict[str, str]] = field(default_factory=lambda: {
        'pubsub.$': {
            "name": "Pubsub Service",
            "category": "pubsub",
            "type": "service",
            "var": "http://jabber.org/protocol/pubsub",
            "extra": {}
        },
        'upload.$': {
            "name": "HTTP File Upload",
            "category": "store",
            "type": "file",
            "var": "urn:xmpp:http:upload:0",
            "extra": {
                "max-size": 5242880
            }
        }
    })



import asyncio
from asyncio import Transport
from enum import Enum
from xml.etree import ElementTree as ET
from xml.sax import ContentHandler

from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.stream.negotiators.StreamNegotiator import StreamNegotiator
from pyjabber.stream.handlers.StanzaHandler import  StanzaHandler
from pyjabber.stream.QueueBridge import QueueBridge
from pyjabber.utils import ClarkNotation as CN


class XMLParser(ContentHandler):
    """
        Manages the stream data and process the XML objects.
        Inheriting from sax.ContentHandler

        :param transport: Transport instance of the connected client. Used to send replays
    """
    __slots__ = ()
    def __init__(self, transport, protocol, stream_negotiator=StreamNegotiator, stanza_handler=StanzaHandler):
        super().__init__()
        self._stack = []

        self._transport = transport
        self._protocol = protocol
        self._stream_negotiator = QueueBridge(
            transport, protocol, self, stream_negotiator, stanza_handler
        )

        self._connection_manager = ConnectionManager()
        self._peer = transport.get_extra_info("peername")

        self._queue_bridge_task = asyncio.create_task(self._stream_negotiator.feed())

    @property
    def transport(self) -> Transport:
        return self._transport

    @transport.setter
    def transport(self, transport: Transport):
        self._transport = transport

    def startElementNS(self, name, qname, attrs):
        if self._stack:  # "<stream:stream>" tag already present in the data stack
            elem = ET.Element(
                CN.clark_from_tuple(name),
                attrib={
                    CN.clark_from_tuple(key): item for key,
                    item in dict(attrs).items()})
            self._stack.append(elem)

        elif name[1] == "stream" and name[0] == "http://etherx.jabber.org/streams":
            elem = ET.Element(
                CN.clark_from_tuple(name),
                attrib={
                    CN.clark_from_tuple(key): item for key,
                    item in dict(attrs).items()})
            self._stack.append(elem)

            self._stream_negotiator.put(elem)

        else:
            raise Exception()

    def endElementNS(self, name, qname):
        if "stream" in name:
            self._connection_manager.close(self._peer)
            self._stack.clear()
            return

        if not self._stack:
            raise Exception()

        elem = self._stack.pop()

        if elem.tag != CN.clark_from_tuple(name):
            # INVALID STANZA/MESSAGE
            raise Exception()

        if self._stack[-1].tag != '{http://etherx.jabber.org/streams}stream':
            self._stack[-1].append(elem)

        else:
            self._stream_negotiator.put(elem)

    def characters(self, content: str) -> None:
        if not self._stack:
            raise Exception()

        elem = self._stack[-1]
        if len(elem) != 0:
            child = elem[-1]
            child.tail = (child.tail or '') + content

        else:
            elem.text = (elem.text or '') + content

    def reset_stack(self) -> None:
        self._stack.clear()

    def cancel_queue_bridge(self):
        self._queue_bridge_task.cancel()



from xml.etree.ElementTree import Element

from pyjabber.stream.utils.Stream import Stream


def test_namespaces_enum():
    assert Stream.Namespaces.XMLSTREAM.value == "http://etherx.jabber.org/streams"
    assert Stream.Namespaces.CLIENT.value == "jabber:client"
    assert Stream.Namespaces.SERVER.value == "jabber:server"


def test_stream_initialization():
    stream = Stream(
        id='12345',
        from_='user@example.com',
        to='protocols@example.com',
        version='1.0',
        xml_lang='en'
    )
    assert isinstance(stream, Element)
    assert stream.tag == 'stream:stream'
    assert stream.attrib['id'] == '12345'
    assert stream.attrib['from'] == 'user@example.com'
    assert stream.attrib['to'] == 'protocols@example.com'
    assert stream.attrib['version'] == '1.0'
    assert stream.attrib['xml:lang'] == 'en'
    assert stream.attrib['xmlns'] == Stream.Namespaces.CLIENT.value
    assert stream.attrib['xmlns:stream'] == Stream.Namespaces.XMLSTREAM.value


def test_stream_open_tag():
    stream = Stream(
        id='12345',
        from_='user@example.com',
        to='protocols@example.com',
        version='1.0',
        xml_lang='en'
    )
    open_tag = stream.open_tag()
    assert open_tag == b"<stream:stream id='12345' from='user@example.com' to='protocols@example.com' version='1.0' xml:lang='en' xmlns='jabber:client' xmlns:stream='http://etherx.jabber.org/streams'>"


def test_response_stream_client():
    attrs = {
        (None, "from"): "user@example.com",
        (None, "to"): "protocols@example.com",
        (None, "version"): "1.0",
        ("http://www.w3.org/XML/1998/namespace", "lang"): "en"
    }
    open_tag = Stream.responseStream(attrs)
    tag = open_tag.decode()

    assert tag.startswith("<stream:stream id='")
    assert "from='protocols@example.com'" in tag
    assert "to='user@example.com'" in tag
    assert "version='1.0'" in tag
    assert "xml:lang='en'" in tag
    assert "xmlns='jabber:client'" in tag
    assert "xmlns:stream='http://etherx.jabber.org/streams'" in tag


def test_response_stream_server():
    attrs = {
        (None, "from"): "protocols.com",
        (None, "to"): "otherser.com",
        (None, "version"): "1.0",
        ("http://www.w3.org/XML/1998/namespace", "lang"): "en"
    }
    open_tag = Stream.responseStream(attrs, True)
    tag = open_tag.decode()

    assert tag.startswith("<stream:stream id='")
    assert "from='otherser.com'" in tag
    assert "to='protocols.com'" in tag
    assert "version='1.0'" in tag
    assert "xml:lang='en'" in tag
    assert "xmlns='jabber:server'" in tag
    assert "xmlns:stream='http://etherx.jabber.org/streams'" in tag



import json
import os
import socket
from unittest.mock import mock_open, patch

from pyjabber.server_parameters import Parameters


def test_default_parameters():
    param = Parameters()
    assert param.host == "localhost"
    assert param.client_port == 5222
    assert param.server_port == 5269
    assert param.server_out_port == 5269
    assert param.family == socket.AF_INET
    assert param.connection_timeout == 60
    assert param.database_path == os.path.join(os.getcwd(), "pyjabber.db")
    assert param.database_purge is False
    assert param.database_in_memory is False
    assert param.cert_path is None
    assert param.message_persistence is True


def test_from_json():
    data_mock = {
        'host': 'publichost',
        'client_port': 5252,
        'server_port': 2525,
        'server_out_port': 6952,
        'family': socket.AF_INET6.value,
        'connection_timeout': 59,
        'database_path': '/some/fake/path',
        'database_purge': True,
        'database_in_memory': True,
        'cert_path': '/another/fake/path',
        'message_persistence': False,
        'verbose': True,
        'plugins': [
            'http://jabber.org/protocol/disco#info',
            'http://jabber.org/protocol/disco#items'
        ],
        'items': {
            'pubsub.$': {
                "name": "Pubsub Service",
                "category": "pubsub",
                "type": "service",
                "var": "http://jabber.org/protocol/pubsub"
            }
        }
    }
    m = mock_open(read_data=json.dumps(data_mock))

    with patch("pyjabber.server_parameters.open", m):
        param = Parameters.from_json("fake.json")

    assert param.host == data_mock["host"]
    assert param.client_port == data_mock["client_port"]
    assert param.server_port == data_mock["server_port"]
    assert param.server_out_port == data_mock["server_out_port"]
    assert param.family == data_mock["family"]
    assert param.connection_timeout == data_mock["connection_timeout"]
    assert param.database_path == data_mock["database_path"]
    assert param.database_purge is data_mock["database_purge"]
    assert param.database_in_memory is data_mock["database_in_memory"]
    assert param.cert_path == data_mock["cert_path"]
    assert param.message_persistence is data_mock["message_persistence"]
    assert param.verbose is True
    assert param.plugins == data_mock["plugins"]
    assert param.items == data_mock["items"]


def test_update_from_json():
    data_mock = {
        'host': 'publichost',
        'client_port':  5252,
        'server_port': 2525,
        'server_out_port': 6952,
        'family': socket.AF_INET6.value,
        'connection_timeout': 59,
        'database_path': '/some/fake/path',
        'database_purge': True,
        'database_in_memory': True,
        'cert_path': '/another/fake/path',
        'message_persistence': False,
        'verbose': True,
        'plugins': [
            'http://jabber.org/protocol/disco#info',
            'http://jabber.org/protocol/disco#items'
        ],
        'items': {
            'pubsub.$': {
                "name": "Pubsub Service",
                "category": "pubsub",
                "type": "service",
                "var": "http://jabber.org/protocol/pubsub"
            }
        }
    }
    m = mock_open(read_data=json.dumps(data_mock))

    with patch("pyjabber.server_parameters.open", m):
        param = Parameters().update_from_json("fake.json")

    assert param.host == data_mock["host"]
    assert param.client_port == data_mock["client_port"]
    assert param.server_port == data_mock["server_port"]
    assert param.server_out_port == data_mock["server_out_port"]
    assert param.family == data_mock["family"]
    assert param.connection_timeout == data_mock["connection_timeout"]
    assert param.database_path == data_mock["database_path"]
    assert param.database_purge is data_mock["database_purge"]
    assert param.database_in_memory is data_mock["database_in_memory"]
    assert param.cert_path == data_mock["cert_path"]
    assert param.message_persistence is data_mock["message_persistence"]
    assert param.verbose is True
    assert param.plugins == data_mock["plugins"]
    assert param.items == data_mock["items"]


def test_pickle():
    filepath = './params.pkl'
    param = Parameters()
    param.dump(filepath)

    assert os.path.exists(filepath)

    parsed_param = Parameters.load(filepath)

    assert parsed_param.host == "localhost"
    assert parsed_param.client_port == 5222
    assert parsed_param.server_port == 5269
    assert parsed_param.server_out_port == 5269
    assert parsed_param.family == socket.AF_INET
    assert parsed_param.connection_timeout == 60
    assert parsed_param.database_path == os.path.join(os.getcwd(), "pyjabber.db")
    assert parsed_param.database_purge is False
    assert parsed_param.database_in_memory is False
    assert parsed_param.cert_path is None
    assert parsed_param.message_persistence is True
    assert parsed_param.plugins == [
        'http://jabber.org/protocol/disco#info',
        'http://jabber.org/protocol/disco#items',
        'http://jabber.org/protocol/pubsub',
        'http://jabber.org/protocol/pubsub#publish',
        'http://jabber.org/protocol/pubsub#subscribe',
        'http://jabber.org/protocol/pubsub#config-node',
        'http://jabber.org/protocol/pubsub#create-nodes',
        'http://jabber.org/protocol/pubsub#delete-nodes',
        'jabber:iq:register',
        'jabber:x:data',
        'urn:xmpp:ping',
        'jabber:iq:rpc'
    ]
    assert parsed_param.items == {
        'pubsub.$': {
            "name": "Pubsub Service",
            "category": "pubsub",
            "type": "service",
            "var": "http://jabber.org/protocol/pubsub"
        }
    }

    if os.path.exists(filepath):
        os.remove(filepath)



import asyncio
from asyncio import Transport
from typing import Union
from xml import sax
from xml.etree.ElementTree import Element
from xml.sax._exceptions import SAXParseException

from loguru import logger

from pyjabber import AppConfig
from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.parsers.XMLParser import XMLParser
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.utils.TransportProxy import TransportProxy
from pyjabber.stream.handlers.ServerStanzaHandler import ServerStanzaHandler
from pyjabber.stream.handlers.StanzaHandler import InternalServerError
from pyjabber.stream.negotiators.ServerIncomingStreamNegotiator import ServerIncomingStreamNegotiator


class XMLProtocol(asyncio.Protocol):
    """
    Protocol to manage the network connection between nodes in the XMPP network. Handles the transport layer.

    :param namespace: namespace of the XML tags (jabber:client or jabber:server)
    :param connection_timeout: Max time without any response from a client. After that, the protocols will terminate the connection
    """
    __slots__ = ('_xmlns', '_host', '_connection_timeout', '_cert_path', '_connection_manager',
                 '_presence_manager', '_tls_queue', '_transport', '_peer', '_xml_parser',
                 '_timeout_monitor', '_timeout_flag', '_connection_type', '_server_log', '_server_incoming')

    def __init__(self, namespace, connection_timeout):
        if namespace not in ["jabber:server", "jabber:client"]:
            raise ValueError('Namespace must be "jabber:server" or "jabber:client')

        self._xmlns = namespace
        self._connection_timeout = connection_timeout

        self._connection_manager = ConnectionManager()
        self._presence_manager = Presence()

        self._transport: Union[Transport, TransportProxy, None] = None
        self._peer = None
        self._xml_parser = None
        self._timeout_monitor = None
        self._timeout_flag = False

        self._server_incoming = namespace == 'jabber:server'

    @property
    def transport(self):
        return self._transport

    @transport.setter
    def transport(self, transport: Union[Transport, TransportProxy]):
        self._transport = transport

    @property
    def namespace(self):
        return self._xmlns

    def connection_made(self, transport):
        """
        Called when a client or another protocols opens a TCP connection to the protocols

        :param transport: The transport object for the connection
        :type transport: asyncio.Transport
        """
        self._peer = transport.get_extra_info('peername')
        logger.info(f"{'Server c' if self._server_incoming else 'C'}onnection from <{self._peer}>")

        if self._connection_timeout:
            self._timeout_monitor = StreamAlivenessMonitor(
                timeout=self._connection_timeout,
                callback=self.connection_timeout
            )

        if AppConfig.app_config.verbose:
            self._transport = TransportProxy(transport, self._peer)
        else:
            self._transport = transport

        self._xml_parser = sax.make_parser()
        self._xml_parser.setFeature(sax.handler.feature_namespaces, True)
        self._xml_parser.setFeature(sax.handler.feature_external_ges, False)

        if self._server_incoming:
            self._xml_parser.setContentHandler(
                XMLParser(
                    self._transport,
                    self,
                    stream_negotiator=ServerIncomingStreamNegotiator,
                    stanza_handler=ServerStanzaHandler
                )
            )

            self._connection_manager.connection_server_incoming(
                self._peer, self._transport
            )
        else:
            self._xml_parser.setContentHandler(
                XMLParser(self._transport, self)
            )

            self._connection_manager.connection(
                self._peer, self._transport
            )



    def connection_lost(self, exc):
        """
        Called when a client or another protocols closes a TCP connection to the protocols

        :param exc: Exception that caused the connection to close
        :type exc: Exception
        """
        if self._timeout_flag:
            return

        logger.info(f"Connection lost {self._peer} {f'Reason: {exc}' if exc else ''}")

        self._transport = None
        self._xml_parser.getContentHandler().cancel_queue_bridge()
        self._xml_parser = None
        self._timeout_monitor.cancel()

        if self._server_incoming:
            host = self._connection_manager.get_host(self._peer)
            self._presence_manager.put_nowait(
                (host, Element("presence", attrib={"type": "INTERNAL"}))
            )
            # self._connection_manager.close_server_incoming(self._peer)
        else:
            jid = self._connection_manager.get_jid(self._peer)
            if jid and jid.user and jid.domain:
                self._presence_manager.put_nowait(
                    (jid, Element("presence", attrib={"type": "INTERNAL"}))
                )
            # self._connection_manager.close(self._peer)

        super().connection_lost(None)


    def data_received(self, data):
        """
        Called when data is received from the client or another protocols

        :param data: Chunk of data received
        :type data: Byte array
        """
        try:
            logger.debug(f"Data received {self._peer}: {data.decode()}")
        except UnicodeDecodeError:
            logger.debug(f"Binary data received {self._peer}")

        if self._timeout_monitor:
            self._timeout_monitor.reset()

        # data = data.replace(b"<?xml version=\'1.0\'?>", b"")
        # data = data.replace(b"<?xml version=\"1.0\"?>", b"")

        try:
            self._xml_parser.feed(data)
        except SAXParseException:
            logger.warning(f"<{self._peer}> sent unparsable data")
            self._transport.close()
        except InternalServerError:
            self._transport.close()

    def eof_received(self):
        """
        Called when the client or another protocols sends an EOF
        """
        if self._transport:
            logger.debug(f"EOF {self._peer}")

    def connection_timeout(self):
        """
        Called when the stream is not responding for a long time
        """
        if not self._transport:
            return

        logger.info(f"Connection timeout {self._peer}")
        if not self._transport.is_closing():
            self._transport.write("<connection-timeout/>".encode())
            self._transport.close()



import os
from aiohttp import web

from pyjabber.webpage.api import api

FILE_PATH = os.path.dirname(os.path.abspath(__file__))

def api_adminpage_app():
    app = web.Application()
    app.router.add_get('/users', api.handleUser)
    app.router.add_get('/roster/{id}', api.handleRoster)
    app.router.add_post('/createuser', api.handleRegister)
    app.router.add_delete('/users/{id}', api.handleDelete)

    return '/api', app

async def get_index(_):
    return web.FileResponse(os.path.join(FILE_PATH, 'build', 'index.html'))

def get_static():
    return os.path.join(FILE_PATH, 'build', 'static')



from asyncio import Transport
from itertools import chain
from typing import List, Optional, Tuple
from uuid import uuid4
from xml.etree import ElementTree as ET

from loguru import logger
from sqlalchemy import and_, delete, insert, select, update

from pyjabber import AppConfig
from pyjabber.db.database import DB
from pyjabber.db.model import Model
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.plugins.xep_0060.enum import (
    Affiliation,
    NodeAttrib,
    SubscribersAttrib,
    Subscription,
)
from pyjabber.plugins.xep_0060.error import ErrorType, error_response
from pyjabber.plugins.xep_0060.utils import success_response
from pyjabber.stanzas.error import StanzaError
from pyjabber.stanzas.Message import Message
from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN
from pyjabber.utils import Singleton


class PubSub(metaclass=Singleton):
    __slots__ = ('_connections', '_jid', '_category', '_var', '_nodes', '_subscribers', '_operations')

    def __init__(self):
        super().__init__()

        self._connections = ConnectionManager()

        pubsub_item = next(
            ((key, item) for key, item
             in AppConfig.app_config.items.items()
             if 'pubsub' in key), None
        )

        self._jid = pubsub_item[0].replace('$', AppConfig.app_config.host)
        self._category = pubsub_item[1]['category']
        self._var = pubsub_item[1]['var']


        self._nodes = None
        self._subscribers = None

        self._operations = {
            'create': self.create_node,
            'delete': self.delete_node,
            'subscribe': self.subscribe,
            'unsubscribe': self.unsubscribe,
            'subscriptions': self.retrieve_subscriptions,
            'items': self.retrieve_items_node,
            'purge': self.purge,
            'publish': self.publish,
            'retract': self.retract
        }

    async def update_memory_from_database(self):
        async with await DB.connection_async() as con:
            query = select(Model.Pubsub)
            res = await con.execute(query)
            self._nodes = res.fetchall()

            query = select(Model.PubsubSubscribers)
            res = await con.execute(query)
            self._subscribers = res.fetchall()

    async def feed(self, jid: JID, element: ET.Element):
        try:
            _, tag = CN.break_down(element[0].tag)

            if tag != 'pubsub':
                return StanzaError.invalid_xml()

            _, operation = CN.break_down(element[0][0].tag)
            return self._operations[operation](element, jid)
        except KeyError as e:
            logger.error(f"Pubsub operation not supported: {e}")
            return StanzaError.feature_not_implemented(feature=str(e), namespace="{http://jabber.org/protocol/pubsub}pubsub")

    def discover_items(self) -> List[tuple]:
        """
        Returns the available nodes at the level specified in the query
        :return: A list of 3-tuples in the format (node, name, type)
        """
        return [
            (node[NodeAttrib.NODE.value], node[NodeAttrib.NAME.value], node[NodeAttrib.TYPE.value])
            for node in self._nodes
        ]

    def discover_info(self, node: str) -> Optional[Tuple[str, str]]:
        """
        Return the info for a given node
        :return: A 2-tuple in the format of (name, type)
        """
        match_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if match_node:
            match_node = match_node.pop()
            return match_node[NodeAttrib.NAME.value], match_node[NodeAttrib.TYPE.value]

        return None

    async def create_node(self, element: ET.Element, jid: JID):
        """
        Creates a new node in the pubsub service.
        The owner will be the creator of the node.
        Configuration can be included in the creation with a form in the IQ request.
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        create = pubsub.find('{http://jabber.org/protocol/pubsub}create')
        config = pubsub.find('{http://jabber.org/protocol/pubsub}config')

        new_node = create.attrib.get('node')
        if not new_node:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        # Node already exists
        if [node for node in self._nodes if node[NodeAttrib.NODE.value] == new_node]:
            return error_response(element, jid, ErrorType.CONFLICT)

        if config:  #pragma: no cover
            pass  # TODO: create node with given configuration

        item = {
            "node": new_node,
            "owner": jid.user,
            "name": None,
            "type": "leaf",
            "max_items": 1024
        }

        async with await DB.connection_async() as con:
            query = insert(Model.Pubsub).values(item)
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        self.update_memory_from_database()

        iq_res, pubsub = success_response(element)
        ET.SubElement(pubsub, 'create', attrib={'node': new_node})
        return ET.tostring(iq_res)

    async def delete_node(self, element: ET.Element, jid: JID):
        """
        Deletes a specific node in the pubsub service.
        ONLY the owner has the permissions to delete.
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub#owner}pubsub')
        delete_stanza = pubsub.find('{http://jabber.org/protocol/pubsub#owner}delete')
        del_node = delete_stanza.attrib.get('node')

        if not del_node:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        node_match = [node for node in self._nodes if node[NodeAttrib.NODE.value] == del_node]
        if not node_match:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        node_match = node_match.pop()
        if node_match[NodeAttrib.OWNER.value] != jid.user:
            return error_response(element, jid, ErrorType.FORBIDDEN)

        async with await DB.connection_async() as con:
            query = delete(Model.Pubsub).where(Model.Pubsub.c.node == del_node)
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

            query = delete(Model.PubsubItems).where(Model.PubsubItems.c.node == del_node)
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        self.update_memory_from_database()
        iq_res, _ = success_response(element)
        return ET.tostring(iq_res)

    async def retrieve_items_node(self, element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        items = pubsub.find('{http://jabber.org/protocol/pubsub}items')
        node = items.attrib.get('node')

        if node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        match_node = [n[NodeAttrib.NODE.value] for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if not match_node:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        target_node: str = match_node.pop()
        is_owner = any(n[NodeAttrib.NODE.value] == target_node and n[NodeAttrib.OWNER.value] == jid.user for n in self._nodes)

        if not is_owner:

            subscribed = any(s[SubscribersAttrib.JID.value] == jid.user
                             and s[SubscribersAttrib.NODE.value] == target_node
                             and s[SubscribersAttrib.SUBSCRIPTION.value] in [Affiliation.MEMBER, Affiliation.PUBLISHER]
                             for s in self._subscribers)

            if not subscribed:
                return error_response(element, jid, ErrorType.FORBIDDEN)

        async with await DB.connection_async() as con:
            query = select(Model.PubsubItems).where(Model.PubsubItems.c.node == target_node)
            res = await con.execute(query)
            res = res.fetchall()

        iq_res, pubsub_res = success_response(element)
        items_res = ET.SubElement(
            pubsub_res,
            '{http://jabber.org/protocol/pubsub}items',
            attrib={"node": target_node}
        )

        for i in res:
            item = ET.SubElement(
                items_res,
                '{http://jabber.org/protocol/pubsub}item', attrib={'id': i[2]}
            )
            item.append(ET.fromstring(i[3]))

        return ET.tostring(iq_res)

    async def subscribe(self, element: ET.Element, jid: JID):
        """
        Subscribe to a specific node
        The default affiliation will be PUBLISHER
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        subscribe = pubsub.find('{http://jabber.org/protocol/pubsub}subscribe')
        node = subscribe.attrib.get('node')
        jid_request = subscribe.attrib.get('jid')

        if node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        if jid_request is None:
            return error_response(element, jid, ErrorType.INVALID_JID)

        jid_request = JID(jid_request)
        if jid_request.bare() != jid.bare():
            return error_response(element, jid, ErrorType.INVALID_JID)

        try:
            target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node][0]
        except IndexError:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        current_state = [
            s for s in self._subscribers
            if s[SubscribersAttrib.JID.value] == jid_request.user and s[SubscribersAttrib.NODE.value] == node
        ]
        if len(current_state) >= 1:
            current_state = current_state.pop()
            if (current_state[SubscribersAttrib.SUBSCRIPTION.value]
               in [Subscription.SUBSCRIBED.value, Subscription.UNCONFIGURED.value]):

                iq_res, pubsub = success_response(element)
                subid = current_state[SubscribersAttrib.SUBID.value]
                ET.SubElement(
                    pubsub,
                    'subscription',
                    attrib={
                        'node': target_node[NodeAttrib.NODE.value],
                        'jid': jid_request.bare(),
                        'subid': subid,
                        'subscription': 'subscribed'
                    }
                )
                return ET.tostring(iq_res)

            elif current_state[SubscribersAttrib.SUBSCRIPTION.value] == Subscription.PENDING.value:
                return error_response(element, jid, ErrorType.PENDING_SUBSCRIPTION)

        subid = str(uuid4())

        item = {
            "node": target_node[NodeAttrib.NODE.value],
            "jid": jid_request.user,
            "subid": subid,
            "subscription": Subscription.SUBSCRIBED.value,
            "affiliation": Affiliation.PUBLISHER
        }

        async with await DB.connection_async() as con:
            query = insert(Model.PubsubSubscribers).values(item)
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        self.update_memory_from_database()

        iq_res, pubsub = success_response(element)
        ET.SubElement(
            pubsub,
            'subscription',
            attrib={
                'node': target_node[NodeAttrib.NODE.value],
                'jid': jid_request.bare(),
                'subid': subid,
                'subscription': 'subscribed'
            }
        )
        return ET.tostring(iq_res)

    async def unsubscribe(self, element: ET.Element, jid: JID):
        """
        Unsubscribe to a specific node
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        unsubscribe = pubsub.find('{http://jabber.org/protocol/pubsub}unsubscribe')
        node = unsubscribe.attrib.get('node')
        jid_request = unsubscribe.attrib.get('jid')
        subid = None

        if node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        if jid_request is None:
            return error_response(element, jid, ErrorType.INVALID_JID)

        jid_request = JID(jid_request)
        if jid_request.bare() != jid.bare():
            return error_response(element, jid, ErrorType.INVALID_JID)

        target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if not target_node:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)
        target_node = target_node.pop()

        number_subscriptions = sum(s[SubscribersAttrib.JID.value] == jid_request.user for s in self._subscribers)

        if number_subscriptions == 0:
            return error_response(element, jid, ErrorType.NOT_SUBSCRIBED)

        if number_subscriptions > 1:
            subid = unsubscribe.attrib.get('subid')
            if subid is None:
                return error_response(element, jid, ErrorType.SUBID_REQUIRED)

            if not any(s[SubscribersAttrib.SUBID.value] == subid for s in self._subscribers):
                return error_response(element, jid, ErrorType.INVALID_SUBID)

            query = delete(Model.PubsubSubscribers).where(
                and_(
                    Model.PubsubSubscribers.c.node == target_node[NodeAttrib.NODE.value],
                    Model.PubsubSubscribers.c.jid == jid_request.user,
                    Model.PubsubSubscribers.c.subid == subid
                )
            )

        else:
            query = delete(Model.PubsubSubscribers).where(
                and_(
                    Model.PubsubSubscribers.c.node == target_node[NodeAttrib.NODE.value],
                    Model.PubsubSubscribers.c.jid == jid_request.user
                )
            )

        async with await DB.connection_async() as con:
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        self.update_memory_from_database()

        iq_res, pubsub = success_response(element)
        sub = ET.SubElement(
            pubsub,
            'subscription',
            attrib={
                'node': target_node[NodeAttrib.NODE.value],
                'jid': jid_request.bare(),
                'subscription': 'none'
            }
        )
        if subid:
            sub.attrib['subid'] = subid
        return ET.tostring(iq_res)

    @staticmethod
    async def retrieve_subscriptions(element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        subscriptions = pubsub.find('{http://jabber.org/protocol/pubsub}subscriptions')
        target_node = subscriptions.attrib.get('node')
        from_stanza = element.attrib.get('from')

        if from_stanza is not None and JID(from_stanza).user != jid.user:
            return error_response(element, jid, ErrorType.FORBIDDEN)

        iq_res, pubsub = success_response(element)
        subscriptions_res = ET.SubElement(pubsub, '{http://jabber.org/protocol/pubsub}subscriptions')

        if target_node is not None and target_node != '':
            query = select(
                Model.PubsubSubscribers.c.node,
                Model.PubsubSubscribers.c.subscription,
                Model.PubsubSubscribers.c.subid
            ).where(
                and_(
                    Model.PubsubSubscribers.c.jid == str(jid.user),
                    Model.PubsubSubscribers.c.node == target_node
                )
            )
        else:
            query = select(
                Model.PubsubSubscribers.c.node,
                Model.PubsubSubscribers.c.subscription,
                Model.PubsubSubscribers.c.subid
            ).where(Model.PubsubSubscribers.c.jid == str(jid.user))

        async with await DB.connection_async() as con:
            res = await con.execute(query)
            res = res.fetchall()

        for sub in res:
            ET.SubElement(subscriptions_res, '{http://jabber.org/protocol/pubsub}subscription', attrib={
                'node': sub[0],
                'jid': jid.bare(),
                'subscription': sub[1],
                'subid': sub[2]
            })

        return ET.tostring(iq_res)

    def retrieve_affiliations(self, element: ET.Element, jid: str):  # pragma: no cover
        pass

    async def purge(self, element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub#owner}pubsub')
        purge = pubsub.find('{http://jabber.org/protocol/pubsub#owner}purge')
        node = purge.attrib.get('node')

        if node is None:
            return error_response(element, jid, ErrorType.NODEID_REQUIRED)

        target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if len(target_node) == 0:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        target_node = target_node.pop()
        if target_node[NodeAttrib.OWNER.value] != jid.user:
            return error_response(element, jid, ErrorType.FORBIDDEN)

        async with await DB.connection_async() as con:
            query = delete(Model.PubsubItems).where(Model.PubsubItems.c.node == node)
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        iq_res, _ = success_response(element, True)
        return ET.tostring(iq_res)

    async def retract(self, element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        retract = pubsub.find('{http://jabber.org/protocol/pubsub}retract')
        node = retract.attrib.get('node')
        item = retract.find('{http://jabber.org/protocol/pubsub}item')
        item_id = item.attrib.get('id') if item is not None else None

        if node is None:
            return error_response(element, jid, ErrorType.NODEID_REQUIRED)

        if item is None or item_id is None:
            return error_response(element, jid, ErrorType.ITEM_REQUIRED)

        target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if len(target_node) == 0:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)
        target_node = target_node.pop()

        current_sub = any(s for s in self._subscribers
                          if s[SubscribersAttrib.JID.value] == jid.user
                          and s[SubscribersAttrib.AFFILIATION.value] == Affiliation.PUBLISHER
                         )

        if jid.user != target_node[NodeAttrib.OWNER.value] and not current_sub:
            return error_response(element, jid, ErrorType.FORBIDDEN)

        async with await DB.connection_async() as con:
            query = delete(Model.PubsubItems).where(
                and_(
                    Model.PubsubItems.c.item_id == item_id,
                    Model.PubsubItems.c.node == node
                )
            )
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        iq_res, pubsub_iq = success_response(element)

        self.send_notification(
            node=target_node[NodeAttrib.NODE.value],
            retract=True,
            item_id=item_id,
            payload=None
        )

        iq_res.remove(pubsub_iq)
        return ET.tostring(iq_res)

    async def publish(self, element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        publish = pubsub.find('{http://jabber.org/protocol/pubsub}publish')
        item = publish.find('{http://jabber.org/protocol/pubsub}item')

        item_id, payload = None, None

        if item is not None:
            item_id = item.attrib.get('id')
            payload = item[0]

        node = publish.attrib.get('node')

        target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if len(target_node) == 0:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        current_sub = [s for s in self._subscribers if s[SubscribersAttrib.AFFILIATION.value] == jid.user]
        if jid.user != target_node[0][NodeAttrib.OWNER.value] \
            or (current_sub and current_sub[0][SubscribersAttrib.AFFILIATION.value] != Affiliation.PUBLISHER):
            return error_response(element, jid, ErrorType.FORBIDDEN)

        if payload is not None:
            async with await DB.connection_async() as con:
                if item_id is not None:
                    query = select(Model.PubsubItems.c.item_id).where(
                        and_(
                            Model.PubsubItems.c.item_id == item_id,
                            Model.PubsubItems.c.node == node
                        )
                    )
                    res = await con.execute(query)
                    res = res.fetchone

                    if res:
                        query = update(Model.PubsubItems).where(
                            and_(
                                Model.PubsubItems.c.node == target_node[NodeAttrib.NODE.value],
                                Model.PubsubItems.c.item_id == item_id
                            )
                        ).values(payload=item_id)

                    else:
                        query = insert(Model.PubsubItems).values({
                            "node": target_node[0][NodeAttrib.NODE.value],
                            "publisher": jid.bare(),
                            "item_id": item_id,
                            "payload": ET.tostring(payload)
                        })

                    await con.execute(query)
                    if not AppConfig.app_config.database_in_memory:
                        await con.commit()

                else:
                    item_id = str(uuid4())
                    query = insert(Model.PubsubItems).values({
                        "node": target_node[0][NodeAttrib.NODE.value],
                        "publisher": jid.bare(),
                        "item_id": item_id,
                        "payload": ET.tostring(payload)
                    })
                    await con.execute(query)
                    if not AppConfig.app_config.database_in_memory:
                        await con.commit()

        self.send_notification(node=target_node[0][NodeAttrib.NODE.value], payload=payload)

        iq_res, pubsub = success_response(element)
        publish = ET.SubElement(pubsub, 'publish', attrib={'node': node})
        if item_id:
            ET.SubElement(publish, 'item', attrib={'id': item_id})
        return ET.tostring(iq_res)

    def send_notification(self, node: str, payload: Optional[ET.Element], item_id: Optional[str] = None, retract: bool = False):
        receivers = [s for s in self._subscribers
                     if s[SubscribersAttrib.NODE.value] == node
                     and s[SubscribersAttrib.AFFILIATION.value] in [Affiliation.MEMBER, Affiliation.PUBLISHER, Affiliation.OWNER]]

        receivers_jid = [r[1] for r in receivers]
        receivers_buffer = [self._connections.get_transport(JID(user=r, domain=AppConfig.app_config.host)) for r in receivers_jid]
        receivers_buffer_single_iterator: List[Tuple[JID, Transport]] = list(chain.from_iterable(receivers_buffer))

        event = ET.Element('event', attrib={'xmlns': 'http://jabber.org/protocol/pubsub#event'})
        items = ET.SubElement(event, 'items', attrib={'node': node})

        if retract:
            retract = ET.SubElement(items, 'retract')
            if item_id:
                retract.attrib['id'] = item_id

        else:
            item = ET.SubElement(items, 'item')
            if item_id:
                item.attrib['id'] = item_id
            if payload:
                item.append(payload)

        for jid, buffer, _ in receivers_buffer_single_iterator:
            message = Message(
                mto=jid.bare(),
                mfrom=AppConfig.app_config.host,
                id=str(uuid4()),
                mtype=None,
                body=event
            )
            buffer.write(ET.tostring(message))



from pyjabber.features.SASL.Mechanism import MECHANISM
from pyjabber.stream.utils.Enums import Signal
from pyjabber.stream.negotiators.StreamNegotiator import Stage, StreamNegotiator


class StreamServerIncomingHandler(StreamNegotiator):
    def __init__(self, transport, starttls, parser_ref) -> None:
        super().__init__(transport, starttls, parser_ref)
        self._ibr_feature = False
        self._sasl_mechanisms = [MECHANISM.EXTERNAL]

        self._stages_handlers[Stage.AUTH] = self._handle_empty_features
        self._stages_handlers.pop(Stage.BIND)

    def _handle_empty_features(self, _):
        self._streamFeature.reset()
        self._transport.write(self._streamFeature.to_bytes())

        self._stage = Stage.READY
        return Signal.DONE




import asyncio
from typing import Union
from uuid import uuid4
from xml.etree import ElementTree as ET

from loguru import logger

from pyjabber import AppConfig
from pyjabber.features.Features import (
    SASL_feature,
    in_band_registration_feature,
    resource_binding_feature,
    start_tls_feature,
    start_tls_proceed_response,
)
from pyjabber.features.SASL.SASL import SASL
from pyjabber.features.StreamFeature import StreamFeature
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.utils.TransportProxy import TransportProxy
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.utils.Enums import Stage, Signal
from pyjabber.stream.handlers.StanzaHandler import InternalServerError
from pyjabber.stream.utils.Stream import Stream
from pyjabber.utils import Exceptions as EX
from pyjabber.utils.Exceptions import NotAuthorizerStreamNegotiationException


class StreamNegotiator:
    __slots__ = ("_host", "_transport", "_peer", "_protocol", "_parser", "_handler", "_stream_feature",
                 "_connection_manager", "_stage", "_ibr_feature", "_sasl", "_stages_handlers")

    def __init__(self, transport, protocol, parser, handler) -> None:
        self._host = AppConfig.app_config.host

        self._transport = transport
        self._peer = transport.get_extra_info("peername")
        self._protocol = protocol
        self._parser = parser
        self._handler = handler

        self._stream_feature = StreamFeature()
        self._connection_manager: ConnectionManager = ConnectionManager()
        self._stage = Stage.CONNECTED

        self._ibr_feature = 'jabber:iq:register' in AppConfig.app_config.plugins

        self._sasl = None

        self._stages_handlers = {
            Stage.CONNECTED: self._handle_init,
            Stage.OPENED: self._handle_tls,
            Stage.SSL: self._handle_init_ssl,
            Stage.SASL: self._handle_ssl,
            Stage.AUTH: self._handle_init_resource_bind,
            Stage.BIND: self._handle_resource_bind
        }

    async def handle_open_stream(self, elem: ET.Element = None) -> Union[Signal, None]:
        try:
            if elem.tag == '{http://etherx.jabber.org/streams}stream':
                self._transport.write(Stream.responseStream(elem.attrib))
            return await self._stages_handlers[self._stage](elem)
        except EX.NotAuthorizerStreamNegotiationException:
            self._transport.write(SE.not_authorized())
            self._connection_manager.close(self._peer)
        except EX.BadRequestException:
            self._transport.write(SE.bad_request())
            self._connection_manager.close(self._peer)
        except InternalServerError:
            self._transport.write(SE.internal_server_error())
            self._connection_manager.close(self._peer)
        except Exception as e:
            logger.error(e)

    async def _handle_init(self, _):
        self._stream_feature.reset()
        self._stream_feature.register(start_tls_feature())
        self._transport.write(self._stream_feature.to_bytes())

        self._stage = Stage.OPENED

    async def _handle_tls(self, element: ET.Element):
        if element.tag == "{urn:ietf:params:xml:ns:xmpp-tls}starttls":
            self._transport.write(start_tls_proceed_response())
            self._transport.pause_reading()

            transport_proxy_is_used = isinstance(self._transport, TransportProxy)

            if transport_proxy_is_used:
                original_transport = self._transport.original_transport
            else:
                original_transport = self._transport

            try:
                loop = asyncio.get_running_loop()
                new_transport = await loop.start_tls(
                    transport=original_transport,
                    protocol=self._protocol,
                    sslcontext=AppConfig.app_config.ssl_context,
                    server_side=True
                )

                if transport_proxy_is_used:
                    new_transport = TransportProxy(new_transport, self._peer)

                self._transport = new_transport
                self._protocol.transport = new_transport
                self._parser.transport = new_transport
                self._handler.transport = new_transport
                self._connection_manager.update_transport_peer(new_transport, self._peer)

                logger.debug(f"Done TLS for <{self._peer}>")
                self._stage = Stage.SSL
                self._parser.reset_stack()
                self._transport.resume_reading()
                return Signal.RESET

            except ConnectionResetError as e:
                logger.error(f"Error during TLS upgrade with <{self._peer}> Reason: {e}")
                self._connection_manager.close(self._peer)
                return Signal.FORCE_CLOSE

        else:
            raise NotAuthorizerStreamNegotiationException()

    async def _handle_init_ssl(self, _):
        self._stream_feature.reset()

        if self._ibr_feature:
            self._stream_feature.register(in_band_registration_feature())

        self._stream_feature.register(SASL_feature())
        self._transport.write(self._stream_feature.to_bytes())

        self._stage = Stage.SASL

    async def _handle_ssl(self, element: ET.Element):
        if self._sasl is None:
            self._sasl = SASL(self._transport, self._parser, self._peer)

        res = await self._sasl.feed(element)

        if res and res == Stage.AUTH:
            self._stage = Stage.AUTH

    async def _handle_init_resource_bind(self, _):
        self._stream_feature.reset()
        self._stream_feature.register(resource_binding_feature())
        self._transport.write(self._stream_feature.to_bytes())

        self._stage = Stage.BIND

    async def _handle_resource_bind(self, element: ET.Element):
        if (element.tag == "{jabber:client}iq"
            and len(element) > 0
            and element[0].tag == "{urn:ietf:params:xml:ns:xmpp-bind}bind"
            and element.attrib.get("type") == "set"):

                resource_id = str(uuid4())

                iq_res = IQ(type_=IQ.TYPE.RESULT, id_=element.get('id') or str(uuid4()))
                bind_res = ET.SubElement(iq_res, "bind", attrib={"xmlns": "urn:ietf:params:xml:ns:xmpp-bind"})

                peername = self._transport.get_extra_info('peername')
                new_jid = self._connection_manager.get_jid(peername)
                new_jid.resource = resource_id

                ET.SubElement(bind_res, 'jid').text = str(new_jid)
                self._transport.write(ET.tostring(iq_res))

                self._connection_manager.set_jid(peername, new_jid, self._transport)
                return Signal.DONE

        else:
            raise NotAuthorizerStreamNegotiationException()



from unittest.mock import patch

import pytest

from pyjabber.server import Server
from pyjabber.server_parameters import Parameters


@pytest.fixture
def setup():
    with patch('pyjabber.protocols.logger') as mock_log, \
         patch('pyjabber.protocols.DB') as mock_db, \
         patch('pyjabber.protocols.init_utils') as mock_utils:
        mock_utils.setup_query_local_ip.return_value = '127.0.0.1'
        # mock_utils.setup_ip_by_host.return_value =
        param = Parameters()
        yield Server(param), mock_log, mock_db


def test_exit_exception():
    with pytest.raises(SystemExit):
        Server.raise_exit()


def test_run_server(setup):
    server, log, db = setup

    # assert log.call_args_list == ["Starting protocols...", "Client domain => localhost"]



import xml.etree.ElementTree as ET

from pyjabber.stream.JID import JID
from pyjabber.stream.handlers.StanzaHandler import StanzaHandler


class StanzaServerIncomingHandler(StanzaHandler):
    def __init__(self, buffer) -> None:
        super().__init__(buffer)

        self._functions = {
            "{jabber:server}iq": self.handle_iq,
            "{jabber:server}message": self.handle_msg,
            "{jabber:server}presence": self.handle_pre
        }

    def handle_iq(self, element: ET.Element):   # pragma: no cover
        return

    def handle_pre(self, element: ET.Element):  # pragma: no cover
        pass

    def handle_msg(self, element: ET.Element):
        jid = JID(element.attrib["to"])

        if not jid.resource:
            priority = self._presenceManager.most_priority(jid)
            if not priority and self._message_persistence:
                self._message_queue.put_nowait(('MESSAGE', jid, ET.tostring(element)))
                return None

            all_resources_online = []
            for user in priority:
                all_resources_online += self._connections.get_transport_online(
                    JID(user=jid.user, domain=jid.domain, resource=user[0]))
            for buffer in all_resources_online:
                buffer[1].write(ET.tostring(element))
        else:
            resource_online = self._connections.get_transport_online(jid)
            if not resource_online and self._message_persistence:
                self._message_queue.put_nowait(('MESSAGE', jid, ET.tostring(element)))
            else:
                for buffer in resource_online:
                    buffer[1].write(ET.tostring(element))



from xml.etree import ElementTree as ET

from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN


class RPC:
    __slots__ = ('_connections',)

    def __init__(self):
        self._connections = ConnectionManager()

    async def feed(self, _, element: ET.Element):
        type_iq = element.attrib.get('type')
        if type_iq == 'set':
            error = self.validate_set_stanza(element)
            if error:
                return SE.not_acceptable(error)

        elif type_iq == 'result':
            error = self.validate_res_stanza(element)
            if error:
                return SE.not_acceptable(error)

        else:
            return SE.invalid_xml()

        to = element.attrib.get('to')
        if not to:
            return SE.invalid_xml()

        buffer = self._connections.get_transport(JID(to))
        for b in buffer:
            b.transport.write(ET.tostring(element))
        return None

    @staticmethod
    def validate_res_stanza(element: ET.Element):
        ns, tag = CN.break_down(element[0].tag)
        if tag != 'query' or ns != 'jabber:iq:rpc':
            return "Malformed response"

        query = element.findall('{jabber:iq:rpc}query')
        if len(query) > 1:
            return "Only 1 query field is permeated"
        method_res = query[0].findall('{jabber:iq:rpc}methodResponse')
        if len(method_res) > 1:
            return "Only 1 methodResponse field is permeated"
        method_name = method_res[0].find('{jabber:iq:rpc}methodName')
        if len(method_name) > 1:
            return "Only 1 methodName field is permeated"
        method_res[0].find('{jabber:iq:rpc}params')

        return None

    @staticmethod
    def validate_set_stanza(element: ET.Element):
        ns, tag = CN.break_down(element[0].tag)
        if tag != 'query' or ns != 'jabber:iq:rpc':
            return "Malformed response"

        query = element.findall('{jabber:iq:rpc}query')
        if len(query) > 1:
            return "Only 1 query field is permeated"
        method_call = query[0].findall('{jabber:iq:rpc}methodCall')
        if len(method_call) > 1:
            return "Only 1 methodCall field is permeated"
        method_name = method_call[0].findall('{jabber:iq:rpc}methodName')
        if len(method_name) > 1:
            return "Only 1 methodName field is permeated"
        method_call[0].findall('{jabber:iq:rpc}params')
        return None

    @staticmethod
    def error_response(to_: JID, from_: JID, id_: str, error: str):
        iq_res = IQ(type_=IQ.TYPE.ERROR, to_=str(to_), from_=str(from_), id_=id_)
        iq_res.append(ET.fromstring(SE.not_acceptable(error)))
        return iq_res



from enum import Enum
from xml.etree import ElementTree as ET

from pyjabber import AppConfig
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID


class ErrorType(Enum):
    INVALID_JID = "<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><invalid-jid xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    CONFLICT = "<error type='auth'><conflict xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    NOT_ACCEPTABLE = "<error type='auth'><not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><nodeid-required xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    ITEM_NOT_FOUND = "<error type='cancel'><item-not-found xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    FORBIDDEN = "<error type='auth'><forbidden xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    PENDING_SUBSCRIPTION = "<error type='auth'><not-authorized xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><pending-subscription xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    SUBID_REQUIRED = "<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><subid-required xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    NODEID_REQUIRED = "<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><nodeid-required xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    ITEM_REQUIRED = "<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><item-required xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    INVALID_SUBID = "<error type='modify'><not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><invalid-subid xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    NOT_SUBSCRIBED = "<error type='cancel'><unexpected-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><not-subscribed xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    NODE_FULL = "<error type='cancel'><conflict xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><node-full xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"
    INVALID_PAYLOAD = "<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><invalid-payload xmlns='http://jabber.org/protocol/pubsub#errors'/></error>"


def error_response(element: ET.Element, jid: JID, error: ErrorType):
    iq_res = IQ(type_=IQ.TYPE.ERROR, from_=AppConfig.app_config.host, to=str(jid), id_=element.attrib.get('id'))
    auth_error = ET.fromstring(error.value)
    iq_res.append(auth_error)
    return ET.tostring(iq_res)



import asyncio
from typing import Dict, List, Union
from uuid import uuid4
from xml.etree import ElementTree as ET

from pyjabber.AppConfig import app_config
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.queues.FailedRemoteConnection import FailedRemoteConnectionWrapper
from pyjabber.queues.NewConnection import NewConnectionWrapper
from pyjabber.queues.PendingMessage import PendingMessageWrapper
from pyjabber.queues.QueueManager import get_queue, QueueName
from pyjabber.stream.JID import JID


async def queue_worker():
    """
       Returns a coroutine that watches two different queues: one
       for new connections and another for new messages.

       - For the new connections queue, the worker checks if there are any pending
         messages in the buffer for the new connection's JID. If so, it processes
         them.
       - For the new messages queue, the worker directly enqueues incoming
         messages into the pending messages buffer.
   """
    local_pending_stanzas: Dict[str, List[bytes]] = {}
    remote_pending_stanzas: Dict[str, List[tuple[JID, bytes]]] = {}

    connection_manager = ConnectionManager()

    con_queue = get_queue(QueueName.CONNECTIONS)
    msg_queue = get_queue(QueueName.MESSAGES)
    s2s_queue = get_queue(QueueName.SERVERS)

    try:
        while True:
            con_task = asyncio.create_task(con_queue.get())
            msg_task = asyncio.create_task(msg_queue.get())

            done, pending = await asyncio.wait(
                [con_task, msg_task],
                return_when=asyncio.FIRST_COMPLETED
            )

            for task in pending:
                task.cancel()

            result: Union[NewConnectionWrapper, PendingMessageWrapper, FailedRemoteConnectionWrapper] = done.pop().result()

            while done:
                item = done.pop().result()
                if isinstance(item, NewConnectionWrapper):
                    await con_queue.put(item)
                else:
                    await msg_queue.put(item)

            if isinstance(result, NewConnectionWrapper):
                if result.client:
                    jid: JID = result.value
                    jid_str = str(jid)
                    jid_bare = jid.bare()

                    target_key = jid_str if jid_str in local_pending_stanzas else (
                        jid_bare if jid_bare in local_pending_stanzas else None)

                    if target_key:
                        stanzas_list = local_pending_stanzas[target_key]
                        buffer = connection_manager.get_transport_online(JID(target_key))

                        while stanzas_list:
                            stanza_bytes = stanzas_list.pop()
                            for b in buffer:
                                b.transport.write(stanza_bytes)
                        local_pending_stanzas.pop(target_key, None)

                else:
                    host = result.value
                    if host in remote_pending_stanzas:
                        stanzas_list = remote_pending_stanzas[host]
                        buffer = connection_manager.get_server_transport_host(host)

                        while remote_pending_stanzas[host]:
                            _, stanza_bytes = stanzas_list.pop()
                            buffer.write(stanza_bytes)
                        remote_pending_stanzas.pop(host, None)

            elif isinstance(result, PendingMessageWrapper):
                jid = result.jid
                payload = result.payload

                if result.is_external:
                    if jid.domain not in remote_pending_stanzas:
                        remote_pending_stanzas[jid.domain] = []
                        await s2s_queue.put(jid.domain)  # Put remote host on connection queue

                    remote_pending_stanzas[jid.domain].append((jid, payload))
                else:
                    if str(jid) not in local_pending_stanzas:
                        local_pending_stanzas[str(jid)] = []
                    local_pending_stanzas[str(jid)].append(payload)

            else:   # FailedRemoteConnectionWrapper
                host = result.value
                error = ET.Element("error", attrib={
                    "type": "cancel" if result.reason == "remote-server-not-found" else "wait"
                })

                if result.reason == "remote-server-not-found":
                    tag = "{urn:ietf:params:xml:ns:xmpp-stanzas}remote-server-not-found"
                else:
                    tag = "{urn:ietf:params:xml:ns:xmpp-stanzas}service-unavailable"

                ET.SubElement(error, tag)

                for jid, payload in remote_pending_stanzas[host]:
                    payload = ET.fromstring(payload)
                    payload.attrib["from"] = app_config.host
                    payload.attrib["to"] = str(jid)
                    payload.attrib["type"] = "error"
                    payload.attrib["id"] = str(uuid4())
                    payload.append(error)

                    for buffer in connection_manager.get_transport(jid):
                        buffer.transport.write(ET.tostring(payload))

                if result.reason == "remote-server-not-found":
                    remote_pending_stanzas.pop(host, None)


    except asyncio.CancelledError:
        pass



from typing import Literal
from xml.etree import ElementTree as ET

from pyjabber import AppConfig
from pyjabber.stanzas.IQ import IQ


def iq_skeleton(element: ET.Element, disco_type: Literal['info', 'items']):
    iq_res = IQ(
        type_=IQ.TYPE.RESULT,
        id_=element.get('id'),
        from_=element.get('to'),
        to=element.get('from')
    )
    return iq_res, ET.SubElement(
        iq_res, f'{{http://jabber.org/protocol/disco#{disco_type}}}query'
    )


def server_info(element: ET.Element):
    iq_res, query = iq_skeleton(element, 'info')
    ET.SubElement(
        query,
        '{http://jabber.org/protocol/disco#info}identity',
        attrib={'category': 'protocols', 'type': 'im', 'name': 'PyJabber Server'})

    for feature in AppConfig.app_config.plugins:
        ET.SubElement(
            query,
            '{http://jabber.org/protocol/disco#info}feature',
            attrib={'var': feature}
        )

    return ET.tostring(iq_res)


def server_items(element: ET.Element):
    iq_res, query = iq_skeleton(element, 'items')
    ET.SubElement(
        query,
        '{http://jabber.org/protocol/disco#items}identity',
        attrib={'category': 'protocols', 'type': 'im', 'name': 'PyJabber Server'})

    for key, info in AppConfig.app_config.items.items():
        jid = key.replace('$', AppConfig.app_config.host) if '$' in key else key
        ET.SubElement(
            query,
            '{http://jabber.org/protocol/disco#items}item',
            attrib={'jid': jid, 'name': info['name']}
        )

    return ET.tostring(iq_res)



import asyncio
import os.path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from pyjabber.network.protocols.XMLProtocol import TransportProxy
from pyjabber.stream.JID import JID
from pyjabber.queues.workers.MessageQueueWorker import queue_worker, tls_worker


@pytest.mark.asyncio
async def test_tls_worker_client():
    with patch('pyjabber.queues.metadata') as mock_metadata, \
         patch('pyjabber.queues.asyncio.get_running_loop') as mock_get_loop, \
         patch('pyjabber.queues.ssl') as mock_ssl:

        mock_queue = MagicMock()
        mock_transport = MagicMock()
        mock_protocol = MagicMock()
        mock_parser = MagicMock()

        mock_loop = MagicMock()
        mock_tls_return = MagicMock()
        mock_loop.start_tls.return_value = mock_tls_return
        mock_get_loop.return_value = mock_loop

        mock_transport.get_extra_info.return_value = ("127.0.0.1", "1234")

        mock_queue.get = AsyncMock()
        mock_queue.get.side_effect = [(mock_transport, mock_protocol, mock_parser), asyncio.CancelledError()]

        mock_metadata.TLS_QUEUE = mock_queue
        mock_AppConfig.app_config.cert_path = os.path.dirname(os.path.abspath(__file__))
        mock_AppConfig.app_config.host = "localhost"

        mock_new_transport = MagicMock()
        mock_loop.start_tls = AsyncMock(return_value=mock_new_transport)

        mock_ssl.create_default_context.return_value = MagicMock()

        try:
            await asyncio.wait_for(tls_worker(), 3)
        except asyncio.TimeoutError:
            pytest.fail()

        mock_loop.start_tls.assert_called_once()
        kwargs = mock_loop.start_tls.call_args.kwargs
        assert kwargs["transport"] == mock_transport.originalTransport
        assert kwargs["protocol"] == mock_protocol
        assert kwargs["sslcontext"].maximum_version == mock_ssl.TLSVersion.TLSv1_2
        assert kwargs["sslcontext"].load_cert_chain.called

        assert type(mock_protocol.transport) == TransportProxy
        assert type(mock_parser.transport) == TransportProxy


@pytest.mark.asyncio
async def test_tls_worker_connection_error_client():
    with patch('pyjabber.queues.ConnectionManager') as mock_con, \
         patch('pyjabber.queues.metadata') as mock_metadata, \
         patch('pyjabber.queues.asyncio.get_running_loop') as mock_get_loop, \
         patch('pyjabber.queues.ssl') as mock_ssl, \
         patch('pyjabber.queues.logger') as mock_logger:

        mock_queue = MagicMock()
        mock_transport = MagicMock()
        mock_protocol = MagicMock()
        mock_parser = MagicMock()

        mock_loop = MagicMock()
        mock_tls_return = MagicMock()
        mock_loop.start_tls.return_value = mock_tls_return
        mock_get_loop.return_value = mock_loop

        mock_peer = ("127.0.0.1", "1234")
        mock_transport.get_extra_info.return_value = mock_peer
        mock_transport.is_closing.return_value = False

        mock_queue.get = AsyncMock()
        mock_queue.get.side_effect = [(mock_transport, mock_protocol, mock_parser), asyncio.CancelledError()]

        mock_metadata.TLS_QUEUE = mock_queue
        mock_AppConfig.app_config.cert_path = ""
        mock_AppConfig.app_config.host = "localhost"

        mock_loop.start_tls = AsyncMock()
        mock_loop.start_tls.side_effect = ConnectionResetError()

        mock_ssl.create_default_context.return_value = MagicMock()

        try:
            await asyncio.wait_for(tls_worker(), 3)
        except asyncio.TimeoutError:
            pytest.fail()

        mock_loop.start_tls.assert_called_once()
        kwargs = mock_loop.start_tls.call_args.kwargs
        assert kwargs["transport"] == mock_transport.originalTransport
        assert kwargs["protocol"] == mock_protocol
        assert kwargs["sslcontext"].maximum_version == mock_ssl.TLSVersion.TLSv1_2
        assert kwargs["sslcontext"].load_cert_chain.called

        mock_logger.error.assert_called_once_with(f"ERROR DURING TLS UPGRADE WITH <{mock_peer}>")
        mock_con.return_value.close.assert_called_once_with(mock_peer)


@pytest.mark.skip
@pytest.mark.asyncio
async def test_tls_worker_connection_queue_client():
    with patch('pyjabber.queues.ConnectionManager') as mock_con, \
         patch('pyjabber.queues.metadata') as mock_metadata, \
         patch('pyjabber.queues.asyncio.get_running_loop') as mock_get_loop, \
         patch('pyjabber.queues.ssl') as mock_ssl, \
         patch('pyjabber.queues.logger') as mock_logger, \
         patch("pyjabber.queues.dict", return_value={"test@localhost": [b"message"]}):
        jid = JID('test@localhost')

        mock_con_queue = MagicMock()
        mock_con_queue.get = AsyncMock()
        mock_con_queue.get.side_effect = [('CONNECTION', jid)]


        mock_metadata.connection_queue.get.return_value = mock_con_queue
        mock_metadata.message_queue.get.return_value = AsyncMock()

        mock_metadata.connection_queue.get.side_effect = [('CONNECTION', jid)]

        await queue_worker()



from xml.etree.ElementTree import Element

from pyjabber.stream.JID import JID


class PluginBase:
    async def feed(self, jid: JID, element: Element):
        """
        The endpoint for any plugin implemented in the protocols
        A plugin may include in the return statement a response message (in bytes) to the client
        """
        raise NotImplementedError



import asyncio
import xml.etree.ElementTree as ET
from asyncio import Transport

from pyjabber import AppConfig
from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.queues.NewConnection import NewConnectionWrapper
from pyjabber.queues.QueueManager import get_queue, QueueName

class ServerStanzaHandler:
    def __init__(self, transport: Transport) -> None:
        self._ip = AppConfig.app_config.ip
        self._transport = transport
        self._peername = transport.get_extra_info('peername')

        self._connections = ConnectionManager()

        self._host = self._connections.get_host(self._peername)

        self._presenceManager = Presence()

        self._message_queue = get_queue(QueueName.MESSAGES)
        self._connection_queue: asyncio.Queue = get_queue(QueueName.CONNECTIONS)

        self._message_persistence = AppConfig.app_config.message_persistence

        self._functions = {
            "{jabber:client}iq": self.handle_iq,
            "{jabber:client}message": self.handle_msg,
            "{jabber:client}presence": self.handle_pre
        }

        get_queue(QueueName.CONNECTIONS).put_nowait(
            NewConnectionWrapper(self._host, False)
        )

    async def feed(self, element: ET.Element):
        pass

    async def handle_iq(self, element: ET.Element):
        pass

    async def handle_msg(self, element: ET.Element):
        pass

    async def handle_pre(self, element: ET.Element):
        pass



from enum import Enum


class Signal(Enum):
    RESET = 0
    STARTTLS = 1
    CLEAR = 2
    DONE = 3
    FORCE_CLOSE = 4


class Stage(Enum):
    """
    Stream connection states.
    """
    CONNECTED = 0
    OPENED = 1
    SSL = 2
    SASL = 3
    AUTH = 4
    BIND = 5
    READY = 6



import re
from asyncio import Transport
from ssl import SSLContext
from typing import List, Optional, Union, NamedTuple
from xml.etree.ElementTree import Element

from loguru import logger

from pyjabber.network.utils.TransportProxy import TransportProxy
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class Peer(NamedTuple):
    ip: str
    port: int


class Client(NamedTuple):
    jid: Optional[JID]
    transport: Transport
    online: bool


class Server(NamedTuple):
    host: str
    transport: Transport


class ConnectionManager(metaclass=Singleton):
    """
    A singleton class used as a repository of connections during the protocols' lifecycle.

    It keeps track of both client and protocols connections using the following data structures:
    """
    def __init__(self) -> None:
        self._peerList: dict[Peer, Client] = {}
        self._remoteList: dict[Peer, Server] = {}
        self._remoteIncomingList: dict[Peer, Server] = {}

        self._orphan_jids: dict[Peer, JID] = {}
        self._orphan_hosts: dict[Peer, str] = {}


    ###########################################################################
    ############################## LOCAL BOUND ################################
    ###########################################################################


    def connection(self, peer: Peer, transport=None) -> None:
        if peer not in self._peerList:
            self._peerList[peer] = Client(None, transport, False)

    def close(self, peer: Peer) -> None:
        """Closes a connection by sending a '</stream:stream> message' and
            deletes it from the peers list
        """
        try:
            client = self._peerList.pop(peer)
            client.transport.write('</stream:stream>'.encode())
            if client.jid:
                self._orphan_jids[peer] = client.jid

            client.transport.close()
        except KeyError:
            logger.error(f"{peer} not present in the peer list")

    def online(self, jid: JID, online: bool = True):
        for peer, client in self._peerList.items():
            if client.jid == jid and client.online != online:
                self._peerList[peer] = self._peerList[peer]._replace(online=online)

    def get_transport(self, jid: JID) -> List[Client]:
        """Get all the available buffers associated with a JID.

                - If the JID is in the full format <username@domain/resource>, it will only return one buffer (or empty list).
                - If the JID is in the bare format <username@domain>, it will return a list of the buffers for each resource available.

            Both cases return a list.
        """
        if jid.resource:
            return [c for c in self._peerList.values() if jid == c.jid]
        else:
            return [c for c in self._peerList.values() if re.match(f"{str(jid)}/*", str(c.jid))]

    def get_transport_online(self, jid: JID) -> List[Client]:
        """Get all the available buffers associated with a JID
            that are ready to receive messages (online).

                - If the JID is in the full format <username@domain/resource>, it will only return one buffer (or empty list).
                - If the JID is in the bare format <username@domain>, it will return a list of the buffers for each resource available.

            Both cases return a list.
        """
        if jid.resource:
            return [c for c in self._peerList.values() if jid == c.jid and c.online]
        else:
            return [c for c in self._peerList.values()
                    if re.match(f"{str(jid)}/*", str(c.jid)) and c.online]

    def update_transport_peer(self, new_transport: Union[Transport, TransportProxy], peer: Peer):
        try:
            self._peerList[peer] = self._peerList[peer]._replace(transport=new_transport)
        except KeyError:
            raise KeyError("Unable to find client with given peer. Check this inconsistency")

    def update_transport_jid(self, new_transport: Union[Transport, TransportProxy], jid: JID):
        if jid.resource is None:
            raise ValueError("JID must have a resource to update transport")

        match = next((peer for peer, client in self._peerList.items()
                      if client.jid == jid), None)
        if match:
            self._peerList[match] = self._peerList[match]._replace(transport=new_transport)
        else:
            raise KeyError("Unable to find client with given JID. Check this inconsistency")

    def get_jid(self, peer: Peer) -> Union[JID, None]:
        try:
            return self._peerList[peer].jid
        except KeyError:
            self._orphan_jids.pop(peer, None)

    def set_jid(self, peer: Peer, jid: JID, transport: Transport = None) -> None:
        """Set/update the jid of a registered connection.

            An optional transport argument can be provided, in order to set/update the stored buffer
        """
        try:
            self._peerList[peer] = self._peerList[peer]._replace(
                jid=jid,
                transport=transport if transport else self._peerList[peer].transport
            )
        except KeyError:
            raise KeyError(f"Unable to find {peer} during jid/transport update")

    def update_resource(self, peer: Peer, resource: str):
        try:
            self._peerList[peer].jid.resource = resource
        except KeyError:
            raise KeyError(f"Unable to find {peer} during resource update")


    ###########################################################################
    ############################# REMOTE SERVER ###############################
    ###########################################################################


    def connection_server(self, peer: Peer, transport: Transport = None, host: str = None) -> None:
        if peer not in self._remoteList:
            self._remoteList[peer] = Server(host, transport)

    def close_server(self, peer: Peer) -> None:
        """Closes a connection by sending a '</stream:stream> message' and
            deletes it from the remote list
        """
        try:
            client = self._remoteList.pop(peer)
            client.transport.write('</stream:stream>'.encode())
            if client.host:
                self._orphan_hosts[peer] = client.host
        except KeyError:
            logger.error(f"{peer} not present in the remote list")

    def get_host(self, peer: Peer) -> Union[str, None]:
        try:
            return self._remoteList[peer].host
        except KeyError:
            return self._orphan_hosts.pop(peer, None)

    def set_host(self, peer: Peer, host: str, transport: Transport = None) -> None:
        """Set/update the host of a registered server connection.

            An optional transport argument can be provided, in order to set/update the stored buffer
        """
        try:
            self._remoteList[peer] = self._remoteList[peer]._replace(
                host=host,
                transport=transport if transport else self._remoteList[peer].transport
            )
        except KeyError:
            raise KeyError(f"Unable to find {peer} during host/transport update")

    def update_host(self, peer: Peer, host: str) -> None:
        try:
            self._remoteList[peer] = self._remoteList[peer]._replace(host=host)
        except KeyError:
            logger.warning("Unable to find protocols with given peer during host update. Check this inconsistency")

    def get_server_transport_peer(self, peer: Peer = None, host: str = None) -> Union[Transport, None]:
        try:
            return self._remoteList.get(peer).transport
        except KeyError:
            return None

    def get_server_transport_host(self, host: str = None) -> Union[Transport, None]:
        return next((s.transport for s in self._remoteList.values() if s.host == host), None)

    def update_transport_server(self, new_transport: Transport, peer: Peer):
        try:
            self._remoteList[peer] = self._remoteList[peer]._replace(transport=new_transport)
        except KeyError:
            logger.warning("Unable to find protocols with given peer. Check this inconsistency")

    def get_server_host(self, peer: Peer) -> Union[str, None]:
        try:
            return self._remoteList.get(peer).host
        except KeyError:
            return None

    def set_host_server(self, peer: Peer, host: str):
        try:
            self._remoteList[peer] = self._remoteList[peer]._replace(host=host)
        except KeyError:
            raise KeyError(f"Unable to find {peer} during host update")

    def get_connection_ssl_certificate(self, peer: Peer) -> Union[SSLContext, None]:
        try:
            self._remoteList.get(peer).transport.get_extra_info("ssl_object")
        except KeyError:
            return None


    ###########################################################################
    ############################# REMOTE SERVER ###############################
    ###########################################################################

    def connection_server_incoming(self, peer: Peer, transport: Transport = None, host: str = None) -> None:
        if peer not in self._remoteIncomingList:
            self._remoteIncomingList[peer] = Server(host, transport)

    def close_server_incoming(self, peer: Peer) -> None:
        """Closes a connection by sending a '</stream:stream> message' and
            deletes it from the remote incoming list
        """
        try:
            client = self._remoteIncomingList.pop(peer)
            client.transport.write('</stream:stream>'.encode())

            if client.host:
                self._presence_manager.put_nowait(
                    (client.host, Element("presence", attrib={"type": "INTERNAL"}))
                )

            client.transport.close()
        except KeyError:
            logger.error(f"{peer} not present in the remote incoming list")



from typing import Tuple
from uuid import uuid4
from xml.etree import ElementTree as ET

from pyjabber import AppConfig
from pyjabber.stanzas.IQ import IQ


def success_response(element: ET.Element, owner: bool = False) -> Tuple[IQ, ET.Element]:
    iq_res = IQ(
        type_=IQ.TYPE.RESULT,
        from_=AppConfig.app_config.host,
        id_=element.attrib.get('id') or str(uuid4())
    )
    if owner:
        xmlns = 'http://jabber.org/protocol/pubsub#owner'
    else:
        xmlns = 'http://jabber.org/protocol/pubsub'

    pubsub = ET.SubElement(iq_res, f'{{{xmlns}}}pubsub')
    return iq_res, pubsub



import asyncio
import ssl
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from socket import AddressFamily
from typing import List, Optional


@dataclass(frozen=True, slots=True)
class AppConfig:
    host: str
    ip: List[str]
    ssl_context: ssl.SSLContext
    ssl_context_s2s: ssl.SSLContext
    connection_timeout: int
    server_port: int
    family: AddressFamily
    config_path: str
    cert_path: str
    root_path:str
    database_path: str
    database_in_memory: bool
    database_purge: bool
    database_debug: bool
    message_persistence: bool
    semaphore: asyncio.Semaphore
    process_pool_exe: ProcessPoolExecutor
    verbose: bool
    plugins: List[str]
    items: dict

app_config: Optional[AppConfig] = None



from typing import Callable, Dict
from xml.etree import ElementTree as ET

from pyjabber import AppConfig
from pyjabber.plugins.xep_0004.field import FieldRequest, FieldTypes
from pyjabber.plugins.xep_0004.xep_0004 import FormType, generate_form
from pyjabber.plugins.xep_0030.utils import iq_skeleton, server_info, server_items
from pyjabber.plugins.xep_0060.xep_0060 import NodeAttrib, PubSub
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class Disco(metaclass=Singleton):
    __slots__ = ('_handlers', '_config_path', '_pubsub', '_pubsub_jid', '_0363_jid', '_0363_max_size')

    def __init__(self):
        self._handlers: Dict[str, Callable[[JID, ET.Element], bytes]] = {
            "info": self.handle_info,
            "items": self.handle_items
        }
        self._config_path: str = AppConfig.app_config.config_path

        if 'http://jabber.org/protocol/pubsub' in AppConfig.app_config.plugins:
            self._pubsub_jid = next((s for s in AppConfig.app_config.items if 'pubsub' in s), None)
            self._pubsub_jid = self._pubsub_jid.replace('$', AppConfig.app_config.host)
            self._pubsub = PubSub()

        if 'urn:xmpp:http:upload:0' in AppConfig.app_config.plugins:
            self._0363_jid = next((s for s in AppConfig.app_config.items if 'upload' in s), None)
            self._0363_max_size = AppConfig.app_config.items[self._0363_jid]["extra"]["max-size"]
            self._0363_jid = self._0363_jid.replace('$', AppConfig.app_config.host)


    async def feed(self, jid: JID, element: ET.Element):
        if len(element) != 1:
            return SE.invalid_xml()

        if element.find('{http://jabber.org/protocol/disco#info}query') is not None:
            return await self._handlers['info'](jid, element)

        elif element.find('{http://jabber.org/protocol/disco#items}query') is not None:
            return await self._handlers['items'](jid, element)

        else:
            return SE.bad_request()

    async def handle_info(self, _, element: ET.Element):
        to = element.attrib.get('to')

        # Pubsub info
        if self._pubsub_jid and to == self._pubsub_jid:
            iq_res, query_res = iq_skeleton(element, 'info')
            query = element.find('{http://jabber.org/protocol/disco#info}query')
            node = query.attrib.get('node')

            if node:
                info = self._pubsub.discover_info(node)

                ET.SubElement(
                    query_res,
                    '{http://jabber.org/protocol/disco#info}identity',
                    attrib={'category': 'pubsub', 'type': info[-1]}
                )
                return ET.tostring(iq_res)

            ET.SubElement(
                query_res,
                '{http://jabber.org/protocol/disco#info}identity',
                attrib={'category': 'pubsub', 'type': 'service'}
            )
            ET.SubElement(
                query_res,
                '{http://jabber.org/protocol/disco#info}feature',
                attrib={'var': 'http://jabber.org/protocol/pubsub'}
            )

            return ET.tostring(iq_res)

        # XEP 0363 (HTTP File Upload) info
        elif self._0363_jid and to == self._0363_jid:
            iq_res, query_res = iq_skeleton(element, 'info')
            ET.SubElement(
                query_res,
                '{http://jabber.org/protocol/disco#info}identity',
                attrib={'category': 'store', 'type': 'file', 'name': 'HTTP File Upload'}
            )
            ET.SubElement(
                query_res,
                '{http://jabber.org/protocol/disco#info}feature',
                attrib={'var': 'urn:xmpp:http:upload:0'}
            )

            dataform = generate_form(
                form_type=FormType.RESULT,
                fields=[
                    FieldRequest(var='FORM_TYPE', field_type=FieldTypes.HIDDEN, values=['urn:xmpp:http:upload:0']),
                    FieldRequest(var='max-file-size', values=[str(self._0363_max_size)])
                ]
            )
            query_res.append(dataform)

            return ET.tostring(iq_res)

        else:
            return server_info(element)

    def handle_items(self, _, element: ET.Element):
        to = element.attrib.get('to')

        if self._pubsub_jid and to == self._pubsub_jid:
            # query = element.find('{http://jabber.org/protocol/disco#items}query')
            # node = query.attrib.get('node')
            items = self._pubsub.discover_items()

            iq_res, query = iq_skeleton(element, 'items')
            for n in items:
                ET.SubElement(
                    query,
                    '{http://jabber.org/protocol/disco#items}item',
                    attrib={
                        'jid': self._pubsub_jid,
                        'node': n[NodeAttrib.NODE.value],
                        'name': n[NodeAttrib.NAME.value]
                })
            return ET.tostring(iq_res)

        else:
            return server_items(element)



import hashlib
import json

from aiohttp import web
from loguru import logger
from sqlalchemy import select, delete, insert

from pyjabber.db.database import DB
from pyjabber.db.model import Model


async def handleUser(_):
    with DB.connection() as con:
        query = select(Model.Credentials.c.id, Model.Credentials.c.jid)
        res = con.execute(query).fetchall()

    users = [{"id": i, "jid": v} for i, v in res]
    return web.Response(text=json.dumps(users))


async def handleRoster(request):
    try:
        user_id = int(request.match_info['id'])

        with DB.connection() as con:
            user_jid = con.execute(
                select(Model.Credentials.c.jid).where(Model.Credentials.c.id == user_id)
            ).fetchone()[0]

            if not user_jid:
                return web.json_response({"status": "error", "message": "User not found"}, status=404)

            roster = con.execute(
                select(Model.Roster.c.roster_item).where(Model.Roster.c.jid.like(f"{user_jid}%"))
            ).fetchall()

            response = [{"item": r[0]} for r in roster]

        return web.Response(text=json.dumps(response))

    except ValueError:
        return web.json_response({"status": "error", "message": "Invalid user ID"}, status=400)
    except Exception as e:
        logger.error(e)
        error_response = {
            "status": "error",
            "message": str(e)
        }
        return web.json_response(error_response, status=500)


async def handleDelete(request):
    try:
        user_id = int(request.match_info['id'])

        with DB.connection() as con:
            res = con.execute(
                select(Model.Credentials.c.id)
            ).fetchall()

            if user_id not in [r[0] for r in res]:
                return web.json_response({"status": "error", "message": "User not found"}, status=404)

            user_jid = con.execute(
                select(Model.Credentials.c.jid).where(Model.Credentials.c.id == user_id)
            ).fetchone()[0]

            con.execute(
                delete(Model.Credentials).where(Model.Credentials.c.id == user_id)
            )
            con.execute(
                delete(Model.Roster).where(Model.Roster.c.jid.like(f"{user_jid}%"))
            )
            con.commit()

        logger.info(f"User with ID {user_id} deleted")
        return web.json_response({"status": "success", "message": "User deleted"}, status=200)

    except ValueError:
        return web.json_response({"status": "error", "message": "Invalid user ID"}, status=400)
    except Exception as e:
        error_response = {
            "status": "error",
            "message": str(e)
        }
        return web.json_response(error_response, status=500)


async def handleRegister(request):
    try:
        data = await request.json()

        with DB.connection() as con:
            res = con.execute(
                select(Model.Credentials.c.jid)
            ).fetchall()

        if any(r[0] == data["jid"] for r in res):
            raise Exception("JID already in use")

        hash_pwd = hashlib.sha256(data["pwd"].encode()).hexdigest()

        with DB.connection() as con:
            con.execute(
                insert(Model.Credentials).values({
                    'jid': data['jid'],
                    'hash_pwd': hash_pwd
                })
            )
            con.commit()

        response_data = {
            "status": "success",
            "message": "OK",
            "received_data": data
        }

        logger.info(f"User with jid {data['jid']} created")
        return web.json_response(response_data, status=200)

    except Exception as e:
        error_response = {
            "status": "error",
            "message": str(e)
        }
        return web.json_response(error_response, status=500)



from xml.etree import ElementTree as ET

from pyjabber import AppConfig
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID


class Ping:
    @staticmethod
    async def feed(jid: JID, element: ET.Element):
        if element.attrib.get('to') == AppConfig.app_config.host:
            return ET.tostring(
                IQ(
                    type_=IQ.TYPE.RESULT,
                    id_=element.attrib.get('id'),
                    from_=AppConfig.app_config.host,
                    to=str(jid)
                )
            )



import asyncio
import xml.etree.ElementTree as ET
from asyncio import Transport

from loguru import logger

from pyjabber import AppConfig
from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.plugins.PluginManager import PluginManager
from pyjabber.queues.NewConnection import NewConnectionWrapper
from pyjabber.queues.PendingMessage import PendingMessageWrapper
from pyjabber.queues.QueueManager import get_queue, QueueName
from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN
from pyjabber.utils.Exceptions import InternalServerError


class StanzaHandler:
    def __init__(self, transport: Transport) -> None:
        self._ip = AppConfig.app_config.ip
        self._transport = transport
        self._peername = transport.get_extra_info('peername')

        self._connections = ConnectionManager()
        self._jid = self._connections.get_jid(self._peername)

        self._pluginManager = PluginManager(self._jid)
        self._presenceManager = Presence()

        self._message_queue = get_queue(QueueName.MESSAGES)
        self._connection_queue: asyncio.Queue = get_queue(QueueName.CONNECTIONS)

        self._message_persistence = AppConfig.app_config.message_persistence

        self._functions = {
            "{jabber:client}iq": self.handle_iq,
            "{jabber:client}message": self.handle_msg,
            "{jabber:client}presence": self.handle_pre
        }

        get_queue(QueueName.CONNECTIONS).put_nowait(
            NewConnectionWrapper(self._jid)
        )

    async def feed(self, element: ET.Element):
        try:
            await self._functions[element.tag](element)
        except (KeyError, InternalServerError) as e:
            logger.error(
                f"Internal protocols error {self._peername}. Closing connection for protocols stability. Reason: ${e}"
            )
            raise InternalServerError

    async def handle_iq(self, element: ET.Element):
        """
            Process the iq stanza with the PluginManager (PM) class

            :param element: The stanza in the ElementTree format
        """
        res = await self._pluginManager.feed(element)
        if res:
            self._transport.write(res)

    async def handle_msg(self, element: ET.Element):
        """
            Router the message to the client

            If the destination client is a user of a remote protocols,
            it will queue the message into the QueueMessage
            object and try to connect to the remote protocols

            :param element: the message in the ElementTree format
        """
        jid = JID(element.attrib.get("to"))

        # Local bound
        if 'from' not in element.attrib:
            element.attrib['from'] = str(self._jid)

        if jid.domain == AppConfig.app_config.host or jid.domain in self._ip:
            if jid.domain in self._ip:
                jid.domain = AppConfig.app_config.host
            if not jid.resource:
                priority = self._presenceManager.most_priority(jid)
                if not priority:
                    if self._message_persistence:
                        await self._message_queue.put(
                            PendingMessageWrapper(jid=JID(jid.bare()), payload=ET.tostring(element))
                        )
                else:
                    all_resources_online = []
                    for user in priority:
                        buffer_online = self._connections.get_transport_online(
                            JID(user=jid.user, domain=jid.domain, resource=user[0])
                        )
                        all_resources_online += buffer_online

                    for buffer in all_resources_online:
                        buffer[1].write(ET.tostring(element))

            else:
                resource_online = self._connections.get_transport_online(jid)
                if not resource_online and self._message_persistence:
                    await self._message_queue.put(
                        PendingMessageWrapper(jid, ET.tostring(element))
                    )
                else:
                    for buffer in resource_online:
                        buffer[1].write(ET.tostring(element))

        # Remote protocols
        else:
            ns, tag = CN.break_down(element.tag)
            if ns == 'jabber:client':
                CN.update_namespace('jabber:server', element)

            buffer = self._connections.get_server_transport_host(jid.domain)
            if buffer:
                buffer.write(ET.tostring(element))
            else:
                await self._message_queue.put(
                    PendingMessageWrapper(jid, ET.tostring(element), external_host=jid.domain)
                )

    async def handle_pre(self, element: ET.Element):
        """
            Handle the presences stanzas
        """
        res = await self._presenceManager.feed(self._jid, element)
        if res:
            self._transport.write(res)



import logging
import os
import socket
import sys

import click
import yaml

from pyjabber.server_parameters import Parameters

if sys.platform != 'win32':
    from uvloop import run
else:
    from winloop import run

from loguru import logger

from pyjabber.server import Server

FILE_PATH = os.path.dirname(os.path.abspath(__file__))


class InterceptHandler(logging.Handler):
    """
    Redirects all logging from stdlib to loguru
    """
    def emit(self, record):
        try:
            lvl = logger.level(record.levelname).name
        except ValueError:
            lvl = record.levelno

        frame, depth = logging.currentframe(), 2
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(lvl, record.getMessage())


def load_config(path=os.path.join(FILE_PATH, "config/config.yaml")):
    try:
        with open(path, 'r') as f:
            raw_config = yaml.safe_load(f) or {}

        flat_config = {}

        for section in ['connection', 'database', 'logs']:
            if section in raw_config and raw_config[section]:
                flat_config.update(raw_config[section])

        for section in ['modules', 'items']:
            if section in raw_config and raw_config[section]:
                flat_config[section] = raw_config[section]

        if 'flags' in raw_config and raw_config['flags']:
            for flag in raw_config['flags']:
                flat_config[flag] = True

        return flat_config
    except FileNotFoundError:
        return {}

config_defaults = load_config()

@click.command(context_settings=dict(default_map=config_defaults))
@click.option('--host', type=str, default='localhost',
              show_default=True, help='Host name')
@click.option('--client_port', type=int, default=5222,
              show_default=True, help='Server-to-client port')
@click.option('--server_port', type=int, default=5269,
              show_default=True, help='Server-to-protocols port')
@click.option('--family',
              type=click.Choice(['ipv4',
                                 'ipv6'],
                                case_sensitive=False),
              default='ipv4',
              show_default=True,
              help='(ipv4 / ipv6)')
@click.option('--timeout', type=int, default=60,
              show_default=True, help='Timeout for connection')
@click.option('--database_path', type=str, default=os.path.join(os.getcwd(), "pyjabber.db"),
              show_default=True, help='Path for database file')
@click.option('--database_purge', is_flag=True, help='Restore database file to default state (empty)')
@click.option('--database_in_memory', is_flag=True,
              help='Database in memory. The data will be erased after protocols shutdown')
@click.option('--message_persistence', is_flag=True,
              help='Keep the unsent messages in memory waiting for the receiver client to connect')
@click.option('--cert_path', type=str, default=None,
              help='Path to the certificate files')
@click.option(
              "-v",
              "--verbose",
              count=True,
              help="Show verbose debug level: -v INFO -vv DEBUG, -vvv level TRACE, ")
@click.option('--log_path', type=str, help='Path to log dumpfile')
@click.option('--debug', '-D', is_flag=True,
              help='Enables debug mode in Asyncio')

def main(
        host,
        client_port,
        server_port,
        family,
        timeout,
        database_path,
        database_purge,
        database_in_memory,
        message_persistence,
        cert_path,
        verbose,
        log_path,
        debug):
    logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)

    logger.remove()

    verbosity = set_verbosity(verbose)

    if log_path:
        log_file = open(os.path.join(log_path, "pyjabber.log"), 'w')
        logger.add(
            log_file,
            enqueue=True,
            format="<green>{time}</green> - <level>{level}: {message}</level>",
            level=verbosity,
        )

    logger.add(
        sys.stderr,
        enqueue=True,
        format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> - <level>{level}: {message}</level>",
        level=verbosity,
    )

    logging.getLogger("sqlalchemy.engine").setLevel(logging.NOTSET)
    logging.getLogger("alembic").setLevel(logging.NOTSET)

    param = Parameters(
        host=host,
        client_port=client_port,
        server_port=server_port,
        family=socket.AF_INET if family == "ipv4" else socket.AF_INET6,
        connection_timeout=timeout,
        database_path=database_path,
        database_purge=database_purge,
        database_in_memory=database_in_memory,
        cert_path=cert_path,
        message_persistence=message_persistence,
        verbose=verbosity == 'TRACE',
        plugins=config_defaults["modules"],
        items=config_defaults["items"],
    )

    server = Server(param)

    run(server.start(), debug=debug)

    return 0


def set_verbosity(verbose):
    if verbose == 1:
        return 'INFO'
    elif verbose == 2:
        return 'DEBUG'
    elif verbose == 3:
        return 'TRACE'
    else:
        return 'SUCCESS'


"""Allow cookiecutter to be executable through `python -m pyjabber`."""

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from typing import List, Optional
from xml.etree import ElementTree as ET

from pyjabber.features.SASL.Mechanism import MECHANISM


def resource_binding_feature():
    return ET.Element("{urn:ietf:params:xml:ns:xmpp-bind}bind")


def in_band_registration_feature():
    return ET.Element("{http://jabber.org/features/iq-register}register")


def SASL_feature(mechanism_list: Optional[List[MECHANISM]]=None):
    if mechanism_list is None:
        mechanism_list = [MECHANISM.PLAIN]

    element = ET.Element("{urn:ietf:params:xml:ns:xmpp-sasl}mechanisms")
    for m in mechanism_list:
        mechanism = ET.SubElement(element, "{urn:ietf:params:xml:ns:xmpp-sasl}mechanism")
        mechanism.text = m.value

    return element

def start_tls_feature(required: bool=True):
    element = ET.Element("{urn:ietf:params:xml:ns:xmpp-tls}starttls")
    if required:
        element.append(ET.Element("required"))

    return element

def start_tls_proceed_response():
    element = ET.Element("{urn:ietf:params:xml:ns:xmpp-tls}proceed")
    return ET.tostring(element)



"""Init

Revision ID: 7ccb2156db5e
Revises: 
Create Date: 2025-05-19 10:18:42.388542

"""
from typing import Sequence, Union

# revision identifiers, used by Alembic.
revision: str = '7ccb2156db5e'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



import xml.etree.ElementTree as ET
from typing import Dict


class Presence(ET.Element):
    def __init__(
            self,
            tag: str = "presence",
            attrib: Dict[str, str] = ...,
            **extra: str) -> None:
        super().__init__(tag, attrib, **extra)



import os
import tempfile
from enum import IntEnum
from typing import Union

import aiofiles
from aiohttp import web
from nanoid import generate

TEMP_DIR = tempfile.mkdtemp(prefix="pyjabber_xep0363_")

class SlotIndex(IntEnum):
    TMPDIR = 0,
    FILENAME = 1,
    CONTENT_TYPE = 2,
    CONTENT_LENGTH = 3,
    IS_UPLOADED_YET = 4

class UploadHttpServer:
    def __init__(self):
        self._slots: dict[str,  # Slot ID
        tuple[
            str,    # Temporal directory to store the file for the slot
            str,    # Filename
            str,    # Content-type (optional)
            int,    # Content-length (optional)
            bool    # Is-uploaded-yet bool
        ]] = {}

        self._root = '/upload'
        self._app = web.Application()
        self._app.router.add_put('/{slot_id}/{file_name}', self.handle_upload)
        self._app.router.add_get('/{slot_id}/{file_name}', self.handle_download)

    @property
    def app(self):
        return self._app

    @property
    def root(self):
        return self._root

    def get_aiohttp_webapp(self) -> tuple[str, web.Application]:
        return self._root, self._app

    def slot_request(self, filename: str, content_type: str = None, content_length: int = None) -> Union[str, None]:
        slot_id = generate(size=8)
        tmpdir = os.path.join(TEMP_DIR, slot_id)

        os.mkdir(tmpdir)
        self._slots[slot_id] = (tmpdir, filename, content_type, content_length, False)

        return slot_id

    async def handle_upload(self, request: web.Request):
        slot_id = request.match_info['slot_id']
        file_name = request.match_info['file_name']
        content_length = request.content_length
        content_type = request.content_type

        if slot_id not in self._slots.keys():
            return web.Response(text=f'{slot_id} not exists. Please first request an available slot via XMPP', status=404)
        if self._slots[slot_id][SlotIndex.IS_UPLOADED_YET.value]:
            return web.Response(text=f'{slot_id} is already filled with a previous upload. Please request a new slot', status=400)
        if file_name != self._slots[slot_id][SlotIndex.FILENAME]:
            return web.Response(text=f'{slot_id} exists, but it is not bound with the file_name: f{file_name}', status=404)

        if (content_length
            and self._slots[slot_id][SlotIndex.CONTENT_LENGTH]
            and content_length != self._slots[slot_id][SlotIndex.CONTENT_LENGTH]
        ):
            return web.Response(text='Invalid content-length', status=400)
        if (content_type
            and self._slots[slot_id][SlotIndex.CONTENT_TYPE]
            and content_type != self._slots[slot_id][SlotIndex.CONTENT_TYPE]
        ):
            return web.Response(text='Invalid file type', status=400)

        save_path = os.path.join(TEMP_DIR, slot_id, file_name)

        expected_size = self._slots[slot_id][SlotIndex.CONTENT_LENGTH]
        total_size = 0
        invalid = False
        async with aiofiles.open(save_path, 'wb') as f:
            async for chunk in request.content.iter_chunked(8192):
                total_size += len(chunk)
                if expected_size and total_size > expected_size:
                    invalid = True
                    break
                await f.write(chunk)

        if invalid or total_size != expected_size:
            os.remove(save_path)
            return web.Response(status=400, text="Invalid file size")

        self._slots[slot_id] = (*self._slots[slot_id][:-1], True)   # type: ignore

        return web.Response(text=f'{file_name} successfully uploaded in {slot_id} slot')

    async def handle_download(self, request: web.Request):
        slot_id = request.match_info['slot_id']
        file_name_req = request.match_info['file_name']

        if slot_id not in self._slots.keys():
            return web.Response(text=f"Slot {slot_id} not Found", status=404)
        if self._slots[slot_id][SlotIndex.IS_UPLOADED_YET] is False:
            return web.Response(text=f"Slot {slot_id} exists, but no file were found", status=204)
        if file_name_req != self._slots[slot_id][SlotIndex.FILENAME]:
            return web.Response(text=f'{slot_id} exists, but it is not bound with the file_name: f{file_name_req}', status=404)

        file_dir = os.path.join(TEMP_DIR, slot_id)
        files = os.listdir(file_dir)
        file_name = files[0]
        file_path = os.path.join(file_dir, file_name)

        return web.FileResponse(
            path=file_path,
            headers={
                "Content-Disposition": f'attachment; filename="{file_name}"'
            }
        )



import socket

from loguru import logger


def setup_query_local_ip():  # pragma: no cover
    """
    Return the local IP of the host machine
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


def setup_ip_by_host(host: str):  # pragma: no cover
    try:
        return socket.gethostbyname(host)
    except socket.gaierror as e:
        logger.error(e)
        return None



from dataclasses import dataclass
from typing import Optional

from pyjabber.stream.JID import JID

@dataclass(frozen=True, slots=True)
class PendingMessageWrapper:
    """
    Represents messages that could not be sent at the time of the request.

    This includes messages for local clients who are currently disconnected,
    as well as messages destined for entities on external servers that not have
    an open connection.
    """
    jid: JID
    payload: bytes
    external_host: Optional[str] = None

    @property
    def is_external(self):
        """
        Indicates whether this message is for an external protocols or not.

        Returns:
            bool: True if this message is for an external protocols or not.
        """
        return self.external_host is not None



from sqlalchemy import Column, Integer, MetaData, String, Table


class Model:
    server_metadata = MetaData()

    Credentials = Table(
        "credentials", server_metadata,
        Column("id", Integer, primary_key=True),
        Column("jid", String, nullable=False),
        Column("hash_pwd", String, nullable=False)
    )

    Roster = Table(
        "roster", server_metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column("jid", String, nullable=False),
        Column("roster_item", String, nullable=False)
    )

    Pubsub = Table(
        "pubsub", server_metadata,
        Column("node", String, primary_key=True),
        Column("owner", String, nullable=False),
        Column("name", String),
        Column("type", String),
        Column("max_items", Integer)
    )

    PubsubSubscribers = Table(
        "pubsub_subscribers", server_metadata,
        Column("node", String, primary_key=True),
        Column("jid", String, primary_key=True),
        Column("subid", String, primary_key=True),
        Column("subscription", String, nullable=False),
        Column("affiliation", String, nullable=False)
    )

    PubsubItems = Table(
        "pubsub_items", server_metadata,
        Column("node", String, primary_key=True),
        Column("publisher", String, nullable=False),
        Column("item_id", String, primary_key=True),
        Column("payload", String, nullable=False)
    )

    PendingSubs = Table(
        "pending_subs", server_metadata,
        Column("jid", String, primary_key=True),
        Column("item", String, primary_key=True)
    )



import asyncio
from typing import Union

from loguru import logger

from pyjabber import AppConfig
from pyjabber.features.Features import start_tls_proceed_response, SASL_feature
from pyjabber.features.SASL.Mechanism import MECHANISM
from pyjabber.features.StreamFeature import StreamFeature
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.utils.TransportProxy import TransportProxy
from pyjabber.stream.negotiators.StreamNegotiator import StreamNegotiator
from pyjabber.stream.utils.Enums import Stage, Signal

from xml.etree import ElementTree as ET

from pyjabber.stream.utils.Stream import Stream
from pyjabber.utils import Exceptions as EX
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stream.handlers.StanzaHandler import InternalServerError
from pyjabber.utils.Exceptions import NotAuthorizerStreamNegotiationException


class ServerIncomingStreamNegotiator(StreamNegotiator):
    def __init__(self, transport, protocol, parser, handler) -> None:
        super().__init__(transport, protocol, parser, handler)
        self._stages_handlers[Stage.AUTH] = self._handle_done_negotiation

    async def handle_open_stream(self, elem: ET.Element = None) -> Union[Signal, None]:
        try:
            if elem.tag == '{http://etherx.jabber.org/streams}stream':
                self._transport.write(Stream.responseStream(elem.attrib))
            return await self._stages_handlers[self._stage](elem)
        except EX.NotAuthorizerStreamNegotiationException:
            self._transport.write(SE.not_authorized())
            self._connection_manager.close_server(self._peer)
        except EX.BadRequestException:
            self._transport.write(SE.bad_request())
            self._connection_manager.close_server(self._peer)
        except InternalServerError:
            self._transport.write(SE.internal_server_error())
            self._connection_manager.close_server(self._peer)
        except Exception as e:
            logger.error(e)

    async def _handle_tls(self, element: ET.Element):
        if element.tag == "{urn:ietf:params:xml:ns:xmpp-tls}starttls":
            self._transport.write(start_tls_proceed_response())
            self._transport.pause_reading()

            transport_proxy_is_used = isinstance(self._transport, TransportProxy)

            if transport_proxy_is_used:
                original_transport = self._transport.original_transport
            else:
                original_transport = self._transport

            try:
                loop = asyncio.get_running_loop()
                new_transport = await loop.start_tls(
                    transport=original_transport,
                    protocol=self._protocol,
                    sslcontext=AppConfig.app_config.ssl_context,
                    server_side=True
                )

                if transport_proxy_is_used:
                    new_transport = TransportProxy(new_transport, self._peer)

                self._transport = new_transport
                self._protocol.transport = new_transport
                self._parser.transport = new_transport
                self._handler.transport = new_transport
                self._connection_manager.update_transport_server(new_transport, self._peer)

                logger.debug(f"Done TLS for <{self._peer}>")
                self._stage = Stage.SSL
                self._parser.reset_stack()
                self._transport.resume_reading()
                return Signal.RESET

            except ConnectionResetError as e:
                logger.error(f"Error during TLS upgrade with <{self._peer}> Reason: {e}")
                self._connection_manager.close_server(self._peer)
                return Signal.FORCE_CLOSE

        else:
            raise NotAuthorizerStreamNegotiationException()

    async def _handle_init_ssl(self, _):
        self._stream_feature.reset()

        self._stream_feature.register(SASL_feature([MECHANISM.EXTERNAL]))
        self._transport.write(self._stream_feature.to_bytes())

        self._stage = Stage.SASL

    async def _handle_done_negotiation(self):
        return Signal.DONE



import logging
import os

import sqlalchemy
from alembic import command
from alembic.config import Config
from loguru import logger
from sqlalchemy import StaticPool, event
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine

from pyjabber import AppConfig
from pyjabber.db.model import Model


class DB:
    _engine = None

    @staticmethod
    def connection() -> sqlalchemy.Connection:  #pragma: no cover
        """
        Returns an already crafted connection with the database.
        It takes the parameters from the protocols class instance (i.e., DB path | DB in memory)
        """
        return DB._engine.connect()

    @staticmethod
    async def connection_async() -> AsyncConnection:  # pragma: no cover
        """
        Returns an already crafted connection with the database.
        It takes the parameters from the protocols class instance (i.e., DB path | DB in memory)
        """
        return DB._engine.connect()

    @staticmethod
    def close_engine():
        """
        Safely dispose the global engine instance used across the protocols
        """
        DB._engine.dispose()

    @staticmethod
    async def close_engine_async():
        """
        Safely dispose the global engine instance used across the protocols
        """
        await DB._engine.dispose()

    @staticmethod
    async def setup_database() -> AsyncEngine:
        """
        Initialize the database that will be used in the protocols session.
        It can be adjusted by the parameters passed in the Server constructor, as it will be
        read via the metadata class
        :return: SQLAlchemy Engine
        """
        if not AppConfig.app_config.database_debug:
            logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)
            logging.getLogger("sqlalchemy.pool").setLevel(logging.WARNING)
            logging.getLogger("sqlite3").setLevel(logging.WARNING)
            logging.getLogger("aiosqlite").setLevel(logging.WARNING)

        if AppConfig.app_config.database_in_memory:
            logger.info("Using database on memory. ANY CHANGE WILL BE LOST AFTER SERVER SHUTDOWN!")
            DB._engine = create_async_engine(
                url="sqlite+aiosqlite:///:memory:",
                isolation_level="AUTOCOMMIT",
                poolclass=StaticPool,
                echo=AppConfig.app_config.database_debug
            )

            @event.listens_for(DB._engine.sync_engine, "connect")
            def set_sqlite_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA synchronous=FULL")
                cursor.execute("PRAGMA journal_mode=MEMORY")
                cursor.close()

            await DB._init_metadata(DB._engine)
            return DB._engine

        if os.path.isfile(AppConfig.app_config.database_path):
            DB._engine = create_async_engine(
                url=f"sqlite+aiosqlite:///{AppConfig.app_config.database_path}",
                echo=AppConfig.app_config.database_debug
            )

        else:
            logger.info("No database found. Initializing one...")
            DB._engine = create_async_engine(
                f"sqlite+aiosqlite:///{AppConfig.app_config.database_path}",
                echo=AppConfig.app_config.database_debug
            )

        @event.listens_for(DB._engine.sync_engine, "connect")
        def set_sqlite_pragma(dbapi_connection, connection_record):
            cursor = dbapi_connection.cursor()
            cursor.execute("PRAGMA journal_mode=WAL")
            cursor.execute("PRAGMA synchronous=NORMAL")
            cursor.execute("PRAGMA cache_size=-20000")
            cursor.close()

        await DB._init_metadata(DB._engine)
        return DB._engine

    @staticmethod
    async def _init_metadata(engine: AsyncEngine):
        def sync_create_all(connection):
            Model.server_metadata.create_all(connection)

        async with engine.begin() as conn:
            await conn.run_sync(sync_create_all)

    @staticmethod
    def run_db_migrations() -> None:
        cfg = Config()
        cfg.set_main_option("script_location", os.path.join(AppConfig.app_config.root_path, '..', 'alembic_local'))
        cfg.set_main_option("sqlalchemy.url", f"sqlite:///{AppConfig.app_config.database_path}")
        command.upgrade(cfg, "head")



import asyncio
import os
from typing import List

from aiohttp import web
from aiohttp.web_app import Application
from loguru import logger

from pyjabber import AppConfig
from pyjabber.webpage.adminPage import get_index, get_static

FILE_PATH = os.path.dirname(os.path.abspath(__file__))

class HttpServer:
    def __init__(self, app_list: List[tuple[str, Application]]):
        self._mainApp = web.Application()
        self._mainApp.router.add_get('/', get_index)
        self._mainApp.router.add_static('/static', get_static())
        for app in app_list:
            self._mainApp.add_subapp(app[0], app[1])


    @property
    def app(self):
        return self._mainApp

    async def start(self):
        runner = web.AppRunner(self._mainApp)
        await runner.setup()
        site = web.TCPSite(runner, AppConfig.app_config.host, 9090)
        try:
            await site.start()
        except OSError as e:
            logger.error(e)
            return

        logger.info(f"Serving admin webpage on http://{AppConfig.app_config.host}:9090")
        try:
            while True:
                await asyncio.sleep(3600)  # Keep alive the protocols
        except asyncio.CancelledError:
            await runner.cleanup()




import asyncio
import socket
from loguru import logger

from pyjabber import AppConfig
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.protocols.XMLProtocolServerOutgoing import XMLProtocolServerOutgoing
from pyjabber.queues.FailedRemoteConnection import FailedRemoteConnectionWrapper
from pyjabber.queues.QueueManager import get_queue, QueueName


async def server_connection_worker():
    """
        Returns a coroutine that watches a queue for connection requests to
        external servers (S2S).
    """
    server_queue = get_queue(QueueName.SERVERS)
    connection_queue = get_queue(QueueName.CONNECTIONS)
    connection_manager = ConnectionManager()
    loop = asyncio.get_running_loop()

    try:
        while True:
            host = await server_queue.get()

            already_open = connection_manager.get_server_transport_host(host)
            if already_open:
                continue

            await loop.create_connection(
                lambda: XMLProtocolServerOutgoing(
                    namespace="jabber:server",
                    host=host,
                    connection_timeout=AppConfig.app_config.connection_timeout,
                ),
                host=host,
                port=AppConfig.app_config.server_port,
                family=AppConfig.app_config.family
            )

    except asyncio.CancelledError:
        pass

    except (ConnectionRefusedError, asyncio.TimeoutError):
        logger.info(f"Remote server <{host}> rejected connection")
        await connection_queue.put(FailedRemoteConnectionWrapper(
            value=host,
            reason="service-unavailable"
        ))

    except socket.gaierror:
        logger.info(f"Remote server <{host}> not found in the DNS lookup")
        await connection_queue.put(FailedRemoteConnectionWrapper(
            value=host,
            reason="remote-server-not-found"
        ))




import asyncio
from asyncio import Transport, Protocol
from xml.etree import ElementTree as ET

from pyjabber.stream.utils.Enums import Signal


class QueueBridge:
    __slots__ = ("_stream_ready", "_transport" ,"_protocol", "_parser" , "_queue", "_stream_handler", "_stanza_handler" , "_stanza_handler_class")

    def __init__(self, transport, protocol, parser, stream_handler, stanza_handler):
        self._stream_ready = False

        self._transport: Transport = transport
        self._protocol: Protocol = protocol
        self._parser = parser

        self._queue = asyncio.Queue()

        self._stream_handler = stream_handler(transport, protocol, parser, self)

        self._stanza_handler_class = stanza_handler
        self._stanza_handler = None

    def put(self, element: ET.Element):
        self._queue.put_nowait(element)

    @property
    def transport(self):
        return self._transport

    @transport.setter
    def transport(self, transport):
        self._transport = transport

    async def feed(self):
        try:
            while True:
                element = await self._queue.get()

                if self._stream_ready:
                    await self._stanza_handler.feed(element)

                else:
                    res = await self._stream_handler.handle_open_stream(element)
                    if res == Signal.RESET:
                        self._parser.reset_stack()
                    if res == Signal.DONE:
                        self._stanza_handler = self._stanza_handler_class(self._transport)
                        self._stream_handler = None
                        self._stream_ready = True

        except asyncio.CancelledError:
            self._transport = None
            self._protocol = None
            self._parser = None
            self._stream_handler = None
            self._stanza_handler = None



from xml.etree import ElementTree as ET

from pyjabber import AppConfig

"""
<stanza-kind from='intended-recipient' to='sender' type='error'>
    [OPTIONAL to include sender XML here]
    <error [by='error-generator'] type='error-type'>
       <defined-condition xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
        [<text xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'
            xml:lang='langcode'>
        OPTIONAL descriptive text
        </text>]
        [OPTIONAL application-specific condition element]
    </error>
</stanza-kind>
"""


XMLNS = "urn:ietf:params:xml:ns:xmpp-stanzas"


def bad_request() -> bytes: # pragma: no cover
    """
   <error type='modify'>
       <bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
   </error>
   """
    return "<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>".encode()


def conflict_error(id: str) -> bytes:
    iq = ET.Element(
        "iq",
        attrib={
            "id": id,
            "type": "error",
            "from": AppConfig.app_config.host})
    error = ET.SubElement(iq, "error", attrib={"type": "cancel"})
    ET.SubElement(
        error, "conflict", attrib={
            "xmlns": "urn:ietf:params:xml:ns:xmpp-stanzas"})
    text = ET.SubElement(
        error, "text", attrib={
            "xmlns": "urn:ietf:params:xml:ns:xmpp-stanzas"})
    text.text = "The requested username already exists"
    return ET.tostring(iq)


def feature_not_implemented(feature: str, namespace: str) -> bytes: # pragma: no cover
    """
    <error type='cancel'>
        <feature-not-implemented
            xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
        <unsupported "
            f"xmlns='<<namespace>>' feature='<<feature>>'/>
    </error>
    """
    return (f"<error type='cancel'><feature-not-implemented xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/><unsupported "
            f"xmlns='{namespace}' feature='{feature}'/></error>").encode()


def invalid_xml() -> bytes: # pragma: no cover
    """
    <stream:error>
        <invalid-xml
            xmlns='urn:ietf:params:xml:ns:xmpp-streams'/>
    </stream:error>
    </stream:stream>
    """
    return f"<stream:error><invalid-xml xmlns='{XMLNS}'/></stream:error></stream:stream>".encode()


def internal_server_error() -> bytes: # pragma: no cover
    """
    <stream:error>
        <internal-protocols-error
            xmlns='urn:ietf:params:xml:ns:xmpp-streams'/>
    </stream:error>
    """
    return "<stream:error><internal-protocols-error xmlns='urn:ietf:params:xml:ns:xmpp-streams'/></stream:error>".encode()

def item_not_found() -> bytes: # pragma: no cover
    """
    <error type='cancel'>
        <item-not-found xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
    </error>
    """
    return f"<error type='cancel'><item-not-found xmlns='{XMLNS}'/></error>".encode()


def not_acceptable(text: str = None) -> bytes:
    """
    <error type='modify'>
        <not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'>
            [OPTIONAL descriptive text]
            <text>
                {ERROR MESSAGE}
            </text>
        </not-acceptable>
    </error>
    """
    if text:
        return f"<error type='modify'><not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'><text>{text}</text></not-acceptable></error>".encode(
        )
    else:
        return "<error type='modify'><not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>".encode()


def not_authorized_sasl() -> bytes: # pragma: no cover
    """
    <failure xmlns='urn:ietf:params:xml:ns:xmpp-sasl'>
        <not-authorized/>
    </failure>
    """
    return "<failure xmlns='urn:ietf:params:xml:ns:xmpp-sasl'><not-authorized/></failure>".encode()


def not_authorized() -> bytes:
    """
    <stream:error>
      <not-authorized xmlns='urn:ietf:params:xml:ns:xmpp-streams'/>
    </stream:error>
    """
    return "<stream:error><not-authorized xmlns='urn:ietf:params:xml:ns:xmpp-streams'/></stream:error>".encode()


def not_well_formed() -> bytes:
    """
    Implicit closes the connection
    """
    return "<stream:error><xml-not-well-formed xmlns='urn:ietf:params:xml:ns:xmpp-streams'/></stream:error></stream:stream>"

def service_unavaliable(): # pragma: no cover
    """
    <error type='cancel'>
        <service-unavailable
            xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
    </error>
    """
    return "<error type='cancel'><service-unavailable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>".encode()



from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class FailedRemoteConnectionWrapper:
    """
    Represents a S2S connection failure.
    Notifies to the client with pending messages to the remote server
    """
    value: str
    reason: str



import enum
import xml.etree.ElementTree as ET
from uuid import uuid4


class Namespaces(enum.Enum):
    """
    Defines the available namespaces in the protocol.
    """
    XMLSTREAM = "http://etherx.jabber.org/streams"
    CLIENT = "jabber:client"
    SERVER = "jabber:server"


class Stream(ET.Element):

    class Namespaces(enum.Enum):
        XMLSTREAM = "http://etherx.jabber.org/streams"
        CLIENT = "jabber:client"
        SERVER = "jabber:server"

    def __init__(
            self,
            id=None,
            from_=None,
            to=None,
            version="1.0",
            xml_lang="en",
            xmlns=Namespaces.CLIENT.value):

        if not id:
            id = str(uuid4())

        attrib = {
            k: v for k, v in (
                ("id", id),
                ("from", from_),
                ("to", to),
                ("version", version),
                ("xml:lang", xml_lang),
                ("xmlns", xmlns)) if v is not None
        }

        attrib["xmlns:stream"] = Namespaces.XMLSTREAM.value

        super().__init__("stream:stream", attrib)

    def open_tag(self) -> bytes:
        tag = f'<{self.tag}'
        for a in self.attrib:
            tag += f" {a}='{self.attrib[a]}'"
        tag += '>'
        return tag.encode()

    @staticmethod
    def responseStream(attrs, server: bool = False):
        attrs = dict(attrs)

        id = str(uuid4())
        from_ = attrs.get((None, "from"))
        to = attrs.get((None, "to"))
        version = attrs.get((None, "version"), "1.0")
        lang = attrs.get(("http://www.w3.org/XML/1998/namespace", "lang"))

        stream = Stream(
            id=id,
            from_=to,
            to=from_,
            version=version,
            xml_lang=lang,
            xmlns=Namespaces.SERVER.value if server else Namespaces.CLIENT.value
        )

        return stream.open_tag()




import asyncio
import multiprocessing
import os
import signal
import ssl
from concurrent.futures.process import ProcessPoolExecutor

from loguru import logger

from pyjabber import AppConfig
from pyjabber.db.database import DB
from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.http_server import HttpServer
from pyjabber.network import CertGenerator
from pyjabber.network.protocols.XMLProtocol import XMLProtocol
from pyjabber.network.protocols.XMLProtocolServerIncoming import XMLProtocolServerIncoming
from pyjabber.plugins.xep_0060.xep_0060 import PubSub
from pyjabber.plugins.xep_0363.upload_server import UploadHttpServer
from pyjabber.plugins.xep_0363.xep_0363 import HTTPFieldUpload
from pyjabber.queues.QueueManager import get_queue, QueueName
from pyjabber.server_parameters import Parameters
from pyjabber.utils.ServerUtils import setup_ip_by_host, setup_query_local_ip
from pyjabber.webpage.adminPage import api_adminpage_app
from pyjabber.queues.workers.MessageQueueWorker import queue_worker
from pyjabber.queues.workers.ServerConnectionWorker import server_connection_worker

SERVER_FILE_PATH = os.path.dirname(os.path.abspath(__file__))

class Server:
    """Server class

    :param param: Instance of Parameter class, with all the configuration available for the protocols. If not provided,
    a default profile will be loaded
    """

    def __init__(self, param: Parameters = Parameters()):
        # Server
        self._host = param.host
        self._public_ip = setup_query_local_ip()
        self._host_ip = setup_ip_by_host(self._host)
        self._client_port = param.client_port
        self._server_port = param.server_port
        self._family = param.family
        self._client_listener = None
        self._server_listener = None
        self._http_server = None
        self._http_apps = []
        self._connection_timeout = param.connection_timeout
        self._database_in_memory = param.database_in_memory

        # Certs
        cert_path = param.cert_path or os.path.join(SERVER_FILE_PATH, "network", "certs")
        try:
            if not CertGenerator.check_hostname_cert_exists(param.host, cert_path):
                CertGenerator.generate_hostname_cert(param.host, cert_path)
        except FileNotFoundError as e:
            logger.error(f"{e.__class__.__name__}: Pass an existing directory in your system to load the certs. "
                         f"Closing protocols")
            raise SystemExit

        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(
            certfile=os.path.join(cert_path, f"{param.host}_cert.pem"),
            keyfile=os.path.join(cert_path, f"{param.host}_key.pem"),
        )

        ssl_context_s2s = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ssl_context_s2s.check_hostname = False
        ssl_context_s2s.verify_mode = ssl.CERT_NONE
        ssl_context_s2s.load_cert_chain(
            certfile=os.path.join(cert_path, f"{param.host}_cert.pem"),
            keyfile=os.path.join(cert_path, f"{param.host}_key.pem"),
        )

        AppConfig.app_config = AppConfig.AppConfig(
            host=param.host,
            ip=[self._host_ip, self._public_ip],
            ssl_context=ssl_context,
            ssl_context_s2s=ssl_context_s2s,
            connection_timeout=param.connection_timeout,
            family=self._family,
            server_port=self._server_port,
            database_path=param.database_path,
            database_purge=param.database_purge,
            database_in_memory=self._database_in_memory,
            database_debug=param.database_debug,
            config_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "config/config.yaml"),
            cert_path=cert_path,
            root_path=SERVER_FILE_PATH,
            message_persistence=param.message_persistence or False,
            semaphore=asyncio.Semaphore(multiprocessing.cpu_count()),
            process_pool_exe=ProcessPoolExecutor(multiprocessing.cpu_count()),
            verbose=param.verbose,
            plugins=param.plugins,
            items=param.items
        )

        # HTTP Server
        if 'urn:xmpp:http:upload:0' in param.plugins and 'upload.$' in param.items:
            self._http_apps.append(api_adminpage_app())
            self._UploadHttpServer = UploadHttpServer()
            self._http_apps.append(self._UploadHttpServer.get_aiohttp_webapp())
            HTTPFieldUpload(self._UploadHttpServer)

            self._http_server = HttpServer(self._http_apps)

        # Flags
        self._ready = asyncio.Event()

    @property
    def ready(self):
        return self._ready

    async def run_server(self):
        """
        Launches the configured protocols, and returns a coroutine.
        """
        try:
            logger.info("Starting protocols...")

            await DB.setup_database()
            if not self._database_in_memory:
                DB.run_db_migrations()

            presence = Presence()
            await presence.get_all_pending_presence()

            pubsub = PubSub()
            await pubsub.update_memory_from_database()

            loop = asyncio.get_running_loop()

            try:
                self._client_listener = await loop.create_server(
                    lambda: XMLProtocol(
                        namespace="jabber:client",
                        connection_timeout=self._connection_timeout,
                    ),
                    host=[self._host, self._public_ip] if self._public_ip else [self._host],
                    port=self._client_port,
                    family=self._family,
                )
            except OSError as e:
                logger.error(e)
                self.raise_exit()

            logger.info(f"Client domain => {self._host}")
            logger.info(
                f"Server is listening clients on {[s.getsockname() for s in self._client_listener.sockets if s]}"
            )

            try:
                self._server_listener = await loop.create_server(
                    lambda: XMLProtocol(
                        namespace="jabber:server",
                        connection_timeout=self._connection_timeout,
                    ),
                    host=[self._host, self._public_ip] if self._public_ip else [self._host],
                    port=self._server_port,
                    family=self._family,
                )
            except OSError as e:
                logger.error(e)
                raise SystemExit

            logger.info(
                f"Server is listening servers on {[s.getsockname() for s in self._server_listener.sockets if s]}"
            )
            logger.success("Server started...")
            self._ready.set()

            while True:
                # Keep the coroutine alive, in order to catch the CancelledError exception
                await asyncio.sleep(3600)

        except asyncio.CancelledError:
            logger.info("Stopping server...")

            await DB.close_engine_async()

            if self._client_listener and self._client_listener.is_serving():
                self._client_listener.close()
                await self._client_listener.wait_closed()

            if self._server_listener and self._server_listener.is_serving():
                self._server_listener.close()
                await self._server_listener.wait_closed()

            logger.success("Server stopped...")

    @staticmethod
    def raise_exit(*args) -> Exception:
        """Exception used to signal the safe close of the server"""
        raise SystemExit(1)

    async def start(self):
        """Start the already created and configuration server"""
        _ = get_queue(QueueName.CONNECTIONS)
        _ = get_queue(QueueName.MESSAGES)
        _ = get_queue(QueueName.SERVERS)

        signal.signal(signal.SIGINT, self.raise_exit)
        signal.signal(signal.SIGABRT, self.raise_exit)
        signal.signal(signal.SIGTERM, self.raise_exit)

        tasks = None

        try:
            tasks = [
                asyncio.create_task(queue_worker()),
                asyncio.create_task(server_connection_worker()),
                asyncio.create_task(self.run_server())
            ]

            if self._http_server:
                tasks.append(
                    asyncio.create_task(self._http_server.start())
                )

            await asyncio.gather(*tasks)

        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError) as e:  # pragma: no cover
            logger.trace(f"Signal {e.__class__.__name__} intercepted. Stopping server")

        finally:
            for task in tasks:
                _ = task.cancel()

            await asyncio.gather(*tasks)
